from ttkbootstrap import Style
from ttkbootstrap.constants import DANGER, DEFAULT
from ttkbootstrap.widgets import Entry

def get_styles() -> Style:
    """Get all defined styling using ttkbootstrap.

    Returns:
        Style: Bootstrap style object
"""
    style = Style()
    style.configure("Error.TEntry",
                    fieldbackground="lightcoral",  # Always red
                    borderwidth=2,
                    relief="solid",
                    foreground="black")
    return style

def highlight_mandatory_entry_on_focus_out(entry: Entry):
    """Highlights a ttk.Entry in red (DANGER) when clicking out of focus if field is empty.

    Args:
        entry (Entry): Entry widget to be modified
    """
    get_styles()
    if entry.get():
        entry.config(bootstyle=DEFAULT)
    else:
        entry.config(style="Error.TEntry")

def highlight_mandatory_entry_on_focus_in(entry: Entry):
    """Changes ttk.Entry back to default settings.

    Args:
        entry (Entry): Entry widget to be modified
    """ 
    entry.config(bootstyle=DEFAULT)

def show_error_in_entry(entry: Entry):
    """Changes ttk.Entry widget to highlight with Bootstrap DANGER style.

    Args:
        entry (Entry): Entry widget to be modified
    """
    get_styles()
    entry.config(style="Error.TEntry")


import tkinter as tk
from ttkbootstrap.icons import Icon
import ttkbootstrap as ttk
from ttkbootstrap.tooltip import ToolTip
from ttkbootstrap.constants import *
from typing import Tuple

class HelpNotification(ttk.Label):
    def __init__(self,parent):
        self.img = tk.PhotoImage(data=Icon.question)
        super().__init__(master=parent,
                        image=self.img,
                        anchor="center")

    def add_help_message(self,msg: str,bootstyle: str | Tuple[str,...]=(SECONDARY,INVERSE)):
        ToolTip(self,
                text=msg,
                bootstyle=bootstyle)



from NekUpload.metadata.metadata import InvenioMetadata 
from NekUpload.upload.invenio_db import InvenioRDM
from NekUpload.metadata.extractor import NekAutoExtractor
from .validate import ValidateSession,ValidateOutput,ValidateGeometry
from typing import List,Optional,Dict
from abc import ABC,abstractmethod
from NekUpload.metadata.relations import Relations,RelationsSchemes,RelationType,ResourceType
from NekUpload.utils.hdf5_reader import HDF5Reader
import os

class NekManager:
    def __init__(self,
                geometry_uploader: 'GeometryManager',
                input_uploader: 'SessionManager',
                output_uploader: 'OutputManager'):

        self.geometry_uploader: GeometryManager = geometry_uploader
        self.input_uploader: SessionManager = input_uploader
        self.output_uploader: OutputManager = output_uploader

        self.auto_metadata_extractor = NekAutoExtractor(
                                    self.input_uploader.session_file,
                                    self.geometry_uploader.geometry_file,
                                    self.output_uploader.output_fld_file)

        self.session_validator = ValidateSession(self.input_uploader.session_file)
        self.geometry_validator = ValidateGeometry(self.geometry_uploader.geometry_file)
        self.output_validator = ValidateOutput(self.output_uploader.output_fld_file)

    def execute_upload(self,url:str,token:str,community_id:str):
        #pass by reference, so all uploaders get updated version of metadata
        self._update_metadata_with_auto_extraction()

        self.geometry_uploader.execute_upload(url,token,community_id)

        #update input and output with geometry upload link
        geometry_doi = self.geometry_uploader.upload_manager.doi
        geometry_record_html = self.geometry_uploader.upload_manager.record_link
        geometry_self_html = self.geometry_uploader.upload_manager.self_link
        geometry_relation = Relations(geometry_doi,RelationsSchemes.DOI,RelationType.CONTINUES,ResourceType.DATASET)
        geometry_relation_html = Relations(geometry_record_html,RelationsSchemes.URL,RelationType.CONTINUES,ResourceType.DATASET)
        geometry_relation_draft_link = Relations(geometry_self_html,RelationsSchemes.URL,RelationType.CONTINUES,ResourceType.DATASET)


        self.input_uploader.metadata_manager.add_related_identifier(geometry_relation)
        self.input_uploader.metadata_manager.add_related_identifier(geometry_relation_html)
        self.input_uploader.metadata_manager.add_related_identifier(geometry_relation_draft_link)
        self.output_uploader.metadata_manager.add_related_identifier(geometry_relation)
        self.output_uploader.metadata_manager.add_related_identifier(geometry_relation_html)
        self.output_uploader.metadata_manager.add_related_identifier(geometry_relation_draft_link)

        self.input_uploader.execute_upload(url,token,community_id)
        self.output_uploader.execute_upload(url,token,community_id)

    def _update_metadata_with_auto_extraction(self):
        results = self.auto_metadata_extractor.extract_data()
        
        if version := results.get("VERSION",None):
            self.geometry_uploader.metadata_manager.add_version(version)
            self.input_uploader.metadata_manager.add_version(version)
            self.output_uploader.metadata_manager.add_version(version)

        #TODO implement everything else

    def validate(self):
        self.session_validator.check_schema()
        self.geometry_validator.check_schema()
        self.output_validator.check_schema()

class UploadManager(ABC):
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def execute_upload(self,url: str,token: str,community_slug: str):
        pass

class GeometryManager(UploadManager):
    def __init__(self,geometry_file: str,
                supporting_files: List[str]=None,
                metadata: InvenioMetadata=None,
                uploader: InvenioRDM=None):
        
        self.geometry_file: str = geometry_file
        self.supporting_files: List[str] = supporting_files if supporting_files else []
        self.metadata_manager = metadata if metadata else InvenioMetadata()
        self.upload_manager = uploader if uploader else InvenioRDM()

    def execute_upload(self,url: str,token: str,community_slug: str):
        files = [self.geometry_file] + self.supporting_files
        metadata_json = self.metadata_manager.get_metadata_payload()
        metadata = {"metadata": metadata_json}

        extra_metadata_file = self.generate_metadata_file()
        files.append(extra_metadata_file)

        try:
            self.upload_manager.upload_files(url,token,files,metadata,community_slug)
        except:
            raise
        finally:
            files.remove(extra_metadata_file)
            os.remove(extra_metadata_file)

    def generate_metadata_file(self) -> str:
        with HDF5Reader(self.geometry_file) as f:
            f.dump_to_plain_file("metadata.rdm")

        return "metadata.rdm"
    
class SessionManager(UploadManager):
    def __init__(self,session_file: str,
                supporting_files: List[str]=None,
                metadata: InvenioMetadata=None,
                uploader: InvenioRDM=None):
        
        self.session_file: str = session_file
        self.supporting_files: List[str] = supporting_files if supporting_files else []
        self.metadata_manager = metadata if metadata else InvenioMetadata()
        self.upload_manager = uploader if uploader else InvenioRDM()

    def execute_upload(self, url, token, community_slug):
        files = [self.session_file] + self.supporting_files
        metadata_json = self.metadata_manager.get_metadata_payload()
        metadata = {"metadata": metadata_json}
        
        self.upload_manager.upload_files(url,token,files,metadata,community_slug)

class OutputManager(UploadManager):
    def __init__(self,output_fld_file: str,
                output_chk_files: List[str]=None,
                filter_files: List[str]=None,
                supporting_files: List[str]=None,
                metadata: InvenioMetadata=None,
                uploader: InvenioRDM=None):
        
        self.output_fld_file: str = output_fld_file
        self.output_chk_files: List[str] = output_chk_files if output_chk_files else []
        self.filter_files: List[str] = filter_files if filter_files else []
        self.supporting_files: List[str] = supporting_files if supporting_files else []
        self.metadata_manager = metadata if metadata else InvenioMetadata()
        self.upload_manager = uploader if uploader else InvenioRDM()

    def execute_upload(self, url, token, community_slug):
        files = [self.output_fld_file] + self.output_chk_files + self.filter_files + self.supporting_files
        metadata_json = self.metadata_manager.get_metadata_payload()
        metadata = {"metadata": metadata_json}

        self.upload_manager.upload_files(url,token,files,metadata,community_slug)


import ttkbootstrap as ttk
from ttkbootstrap.constants import *
from ttkbootstrap.scrolled import ScrolledFrame

class InfoScene(ScrolledFrame):
    def __init__(self, parent):
        super().__init__(parent, autohide=True)

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)

        self.LEFT_MARGIN = 5

        self._add_about_section()
        self._add_capabilities_section()
        self._add_getting_started_section()

        self.bind("<Configure>", self.update_wraplength)

    def update_wraplength(self, event):
        # Dynamically set the wraplength based on the width of the parent frame
        # Subtract a little for padding and margin
        self.about_description.config(wraplength=event.width - 20)
        self.capability_description.config(wraplength=event.width - 20)
        self.get_started_description.config(wraplength=event.width - 20)

    def _add_about_section(self):
        # Create the label for the title
        about_label = ttk.Label(
            master=self,
            text="What is NekUpload?",
            font=("TkDefaultFont", 20, "bold", "underline"),
            anchor="center",
            bootstyle=PRIMARY
        )
        about_label.grid(row=0, column=0, pady=5, padx=self.LEFT_MARGIN, sticky=W)

        # Create the description label
        self.about_description = ttk.Label(
            master=self,
            text=("NekUpload is designed to enforce a consistent upload format for Nektar++ datasets. "
                "This program removes the burden of data management from the user, ensuring files are uploaded in a defined format. "
                "This ensures that data management best practices are enforced, and that users can easily find their files, "
                "which are hosted on an online repository. This also ensures that the online post-processing programs can function correctly."),
            font=("TKDefaultFont", 12),
            anchor="w",
            justify="left",
        )
        self.about_description.grid(row=1, column=0, padx=self.LEFT_MARGIN, pady=10, sticky="nsew")

    def _add_capabilities_section(self):
        # Create the label for the title
        capability_label = ttk.Label(
            master=self,
            text="Capabilities",
            font=("TkDefaultFont", 20, "bold", "underline"),
            anchor="center",
            bootstyle=SECONDARY
        )
        capability_label.grid(row=2, column=0, pady=5, padx=self.LEFT_MARGIN, sticky=W)

        # Create the description label
        self.capability_description = ttk.Label(
            master=self,
            text=("This program currently contains the following capabilities:\n\n"
                " - Upload Nektar++ datasets\n"
                " - Validate datasets are self-consistent and in correct format\n"
                " - Auto-extraction of existing metadata in file to annotate the datasets\n"
                " - Linking of datasets with pre-existing datasets on the database"),
            font=("TKDefaultFont", 12),
            anchor="w",
            justify="left",
        )
        self.capability_description.grid(row=3, column=0, padx=self.LEFT_MARGIN, pady=10, sticky="nsew")

    def _add_getting_started_section(self):
        get_started_label = ttk.Label(
            master=self,
            text="Getting Started",
            font=("TkDefaultFont", 20, "bold", "underline"),
            anchor="center",
            bootstyle=PRIMARY
        )
        get_started_label.grid(row=4, column=0, pady=5, padx=self.LEFT_MARGIN, sticky=W)

        # Create the description label
        self.get_started_description = ttk.Label(
            master=self,
            text=("To get started with this program, first you must know the host name of the online repository and have an API token"
                "This program currently supports two defaults: one for a demo instance of InvenioRDM and another for the AE Datastore. "
                "Please configure your settings before proceeding."),
            font=("TKDefaultFont", 12),
            anchor="w",
            justify="left",
        )
        self.get_started_description.grid(row=5, column=0, padx=self.LEFT_MARGIN, pady=10, sticky="nsew")



import ttkbootstrap as ttk
from ttkbootstrap.constants import *
from NekUpload.frontend.components.header import Header
from NekUpload.frontend.components.menu import Menu
from NekUpload.frontend.scenes.info import InfoScene
from NekUpload.frontend.scenes.upload import UploadScene
from NekUpload.frontend.scenes.review import ReviewScene
from NekUpload.frontend.scenes.explore import ExploreScene
from NekUpload.frontend.scenes.help import HelpScene
from NekUpload.frontend.scenes.settings import SettingScene
from NekUpload.frontend.components.terminal import TerminalHandler,TerminalWidget
from NekUpload.frontend.components.settings_manager import SettingsManager
import logging
import platform
import psutil
import os
class NekUploadNewGUI:
    def __init__(self):
        self.root = ttk.Window(themename="sandstone")
        self.root.title("NekUpload")
        if platform.system() == 'Windows':
            self.root.state('zoomed')  # Maximize on Windows
        else:
            self.root.attributes('-zoomed', True)  # Maximize on macOS/Linux
            
        self.root.rowconfigure(0, weight=0)
        self.root.rowconfigure(1, weight=1)
        self.root.columnconfigure(0, weight=0)
        self.root.columnconfigure(1, weight=20)

        setting_manager = SettingsManager("","","")

        #create a header
        self.header: ttk.Frame = Header(self.root)
        self.header.grid(row=0, column=0, columnspan=2, sticky=(N, E, S, W), ipadx=10, ipady=10)

        #create a menu
        self.menu = Menu(self.root)
        self.menu.grid(row=1, column=0, sticky=NSEW, ipadx=10, ipady=10)

        #pages and which button they link to
        info_page = InfoScene(self.root)
        review_page = ReviewScene(self.root)
        explore_page = ExploreScene(self.root)
        help_page = HelpScene(self.root)
        setting_page = SettingScene(self.root,setting_manager)
        upload_page = UploadScene(self.root,self.root,setting_manager)

        #create page frame and default to INFO
        #menu buttons will direct pages to here
        self.page = info_page
        self.page.grid(row=1, column=1,sticky=NSEW,ipadx=20,ipady=20)  # Show new page

        self.menu.add_link_to_button("UPLOAD",lambda: self.switch_page(upload_page))
        self.menu.add_link_to_button("INFO",lambda: self.switch_page(info_page))
        self.menu.add_link_to_button("REVIEW",lambda: self.switch_page(review_page))
        self.menu.add_link_to_button("EXPLORE",lambda: self.switch_page(explore_page))
        self.menu.add_link_to_button("HELP",lambda: self.switch_page(help_page))
        self.menu.add_link_to_button("SETTINGS",lambda: self.switch_page(setting_page))

        self.terminal = TerminalWidget(self.root)
        self.terminal.grid(row=2,column=0,columnspan=2,sticky=(E,W))

        #configure terminal for logging        
        logger = logging.getLogger()  # Get the root logger
        logger.setLevel(logging.INFO)
        terminal_handler = TerminalHandler(self.terminal)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s') #format of message
        terminal_handler.setFormatter(formatter)
        logger.addHandler(terminal_handler)

    def switch_page(self, new_page: ttk.Frame):
        if self.page != new_page:
            self.page.grid_forget()
            self.page = new_page
            self.page.grid(row=1, column=1,sticky=NSEW,ipadx=20,ipady=20)  # Show new page

    def run(self):
        #show live memory footprint
        def get_memory_usage():
            process = psutil.Process(os.getpid())
            memory_info = process.memory_info()
            return memory_info.rss / (1024 * 1024)  # Return memory usage in MB

        def update_memory_label():
            memory_usage = get_memory_usage()
            memory_label.config(text=f"Memory Usage: {memory_usage:.2f} MB")
            self.root.after(1000, update_memory_label) # Update every 1 second

        memory_label = ttk.Label(self.root, text="Memory Usage: 0.00 MB")
        memory_label.grid(row=10,column=0)

        update_memory_label()

        self.root.mainloop()

def main():
    app = NekUploadNewGUI()
    app.run()

if __name__ == "__main__":
    main()


from enum import Enum
from types import MappingProxyType

#Enums describing various InvenioRDM data types

class RelationsSchemes(Enum):
    DOI = "doi"
    URL = "url"

RELATIONS_SCHEMES_REVERSE_MAP: MappingProxyType[str,RelationsSchemes] = MappingProxyType({
    member.value: member for member in RelationsSchemes
})

class RelationType(Enum):
    CONTINUES = "continues"

RELATIONS_TYPE_REVERSE_MAP: MappingProxyType[str,RelationType] = MappingProxyType({
    "continues": RelationType.CONTINUES
})

class ResourceType(Enum):
    DATASET = "dataset"
    PHYSICAL_OBJECT = "physicalobject"

RESOURCE_TYPE_REVERSE_MAP: MappingProxyType[str,ResourceType] = MappingProxyType({
    "dataset": ResourceType.DATASET,
    "physical_object": ResourceType.PHYSICAL_OBJECT
})


from types import MappingProxyType
from typing import Dict,Any
from .data_types import RelationsSchemes,RelationType,ResourceType,RELATIONS_SCHEMES_REVERSE_MAP,RELATIONS_TYPE_REVERSE_MAP,RESOURCE_TYPE_REVERSE_MAP

RELATION_TYPE_TITLE: MappingProxyType[RelationType,Dict[str,str]] = MappingProxyType({
    RelationType.CONTINUES: {"en": "Continues"}
})

RESOURCE_TYPE_TITLE: MappingProxyType[ResourceType,Dict[str,str]] = MappingProxyType({
    ResourceType.DATASET: {"en": "Dataset"},
    ResourceType.PHYSICAL_OBJECT: {"en": "Physical object"}
})

class Relations:
    """Metadata model for handling the Relations field in InvenioRDM. Describes any links between records with each other and/or external resources
    """
    def __init__(self,
                id: str,
                scheme: RelationsSchemes,
                relation: RelationType,
                resource: ResourceType):
        """Class initialiser

        Args:
            id (str): ID of the resource to be linked
            scheme (RelationsSchemes): ID scheme of the resource to be linked
            relation (RelationType): Describes the relationship of the link
            resource (ResourceType): Describes the resource type of the linked resource
        """
        self.id: str = id
        self.scheme: RelationsSchemes = scheme
        self.relation_type: RelationType = relation
        self.resource_type: ResourceType = resource

    def __eq__(self, other):
        if not isinstance(other, Relations):
            return NotImplemented
        return (self.id == other.id and
                self.scheme == other.scheme and
                self.relation_type == other.relation_type and
                self.resource_type == other.resource_type)

    def to_json(self):
        return {
            "identifier": self.id,
            "scheme": self.scheme.value,
            "relation_type": {
                "id": self.relation_type.value,
                "title": RELATION_TYPE_TITLE[self.relation_type]
            },
            "resource_type": {
                "id": self.resource_type.value,
                "title": RESOURCE_TYPE_TITLE[self.resource_type]
            }
        }
    
    @classmethod
    def from_json(cls,data: Dict[str,Any]) -> 'Relations':
        return Relations(data["identifier"],
                        RELATIONS_SCHEMES_REVERSE_MAP.get(data["scheme"]),
                        RELATIONS_TYPE_REVERSE_MAP.get(data["relation_type"]["id"]),
                        RESOURCE_TYPE_REVERSE_MAP.get(data["resource_type"]["id"]))


__version__="0.1.0"

import logging
import os

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s'
)

#also read logs to a file
file_handler = logging.FileHandler("info.log", mode='w')
file_handler.setLevel(logging.INFO)
file_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))

logging.getLogger().addHandler(file_handler)

LOG_FILE_PATH = os.path.abspath("info.log")
MODULE_ABS_PATH = os.path.dirname(os.path.abspath(__file__))


import ttkbootstrap as ttk
from ttkbootstrap.constants import *
import tkinter as tk
from tkinter import filedialog
from NekUpload.frontend.components.help import HelpNotification
from NekUpload.frontend import style_guide

class UploadSessionFrame(ttk.LabelFrame):
    def __init__(self,parent):
        super().__init__(
            master=parent,
            text="Session",
            bootstyle=DANGER,
            padding=10
            )

        self.columnconfigure(0,weight=1)
        self.columnconfigure(1,weight=1)
        self.rowconfigure(0,weight=1)
        self.rowconfigure(1,weight=1)

        label = ttk.Label(
            master=self,
            text="Upload all input related files here.",
            font=("TKDefaultFont", 12),
            anchor="w",
            justify="left",
        )
        label.grid(row=0,column=0,columnspan=2,pady=2,sticky=(NSEW))

        n = ttk.Notebook(self,bootstyle=SECONDARY)
        n.grid(row=1,column=0,columnspan=2,sticky=NSEW)
        
        f1: ttk.Frame = self._upload_session(n)

        n.add(f1,text="Upload from Source")
    
    def _upload_session(self,parent) -> ttk.Frame:
        frame = ttk.Frame(master=parent)

        frame.rowconfigure(0,weight=1)
        frame.rowconfigure(1,weight=1)
        frame.columnconfigure(0,weight=1)
        frame.columnconfigure(1,weight=1)
        frame.columnconfigure(2,weight=1)

        session_file_frame: ttk.Labelframe = self._session_mandatory_info(frame)
        session_file_frame.grid(row=0,column=0,sticky=NSEW)

        return frame
    
    def _session_mandatory_info(self,parent) -> ttk.Labelframe:
        session_file_frame= ttk.Labelframe(
            master=parent,
            text="Mandatory"
        )
        session_file_frame.columnconfigure(0,weight=1)
        session_file_frame.columnconfigure(1,weight=3)
        session_file_frame.columnconfigure(2,weight=1)
        session_file_frame.columnconfigure(3,weight=1)

        ################################
        # Ask for title
        title_label = ttk.Label(
            master=session_file_frame,
            text="Session Title: "
        )
        title_label.grid(row=0,column=0,sticky=NSEW)
        
        self._session_title = tk.StringVar()
        self.session_title_entry = ttk.Entry(
            master=session_file_frame,
            textvariable=self._session_title,
        )
        self.session_title_entry.grid(row=0,column=1,sticky=NSEW)

        #for highlighting missing fields and turning off red when user is inputting
        self.session_title_entry.bind("<FocusOut>",lambda event: style_guide.highlight_mandatory_entry_on_focus_out(self.session_title_entry))
        self.session_title_entry.bind("<FocusIn>",lambda event: style_guide.highlight_mandatory_entry_on_focus_in(self.session_title_entry))

        ################################
        # Ask for session file
        session_file_label = ttk.Label(
            master=session_file_frame,
            text="Select Session File: "
        )
        session_file_label.grid(row=1,column=0,sticky=NSEW)

        self._session_file = tk.StringVar()
        self.session_file_entry = ttk.Entry(
            master=session_file_frame,
            textvariable=self._session_file,
        )
        self.session_file_entry.grid(row=1,column=1,sticky=NSEW)

        #for highlighting missing fields and turning off red when user is inputting
        self.session_file_entry.bind("<FocusOut>",lambda event: style_guide.highlight_mandatory_entry_on_focus_out(self.session_file_entry))
        self.session_file_entry.bind("<FocusIn>",lambda event: style_guide.highlight_mandatory_entry_on_focus_in(self.session_file_entry))

        def browse_file():
            file_path = filedialog.askopenfilename(
                title="Select Session File",
                filetypes=(("Session Files",".xml"),)
            )
            self._session_file.set(file_path)

        browse = ttk.Button(
            master=session_file_frame,
            text="Browse Files",
            command=browse_file
        )
        browse.grid(row=1,column=2,sticky=NSEW)

        help_logo:ttk.Label = HelpNotification(session_file_frame)
        help_logo.add_help_message("This will be the title of the session record")
        help_logo2:ttk.Label = HelpNotification(session_file_frame)
        help_logo2.add_help_message("Only XML formatted session files are accepted")
        help_logo.grid(row=0, column=3, sticky=NSEW)
        help_logo2.grid(row=1, column=3, sticky=NSEW)

        return session_file_frame
    
    @property
    def session_file_name(self):
        return self._session_file.get()
    
    @property
    def session_dataset_title(self):
        return self._session_title.get()
    
    def add_error_style_to_mandatory_entries(self):
        if not self.session_file_entry.get():
            print("H")
            style_guide.show_error_in_entry(self.session_file_entry)

        if not self.session_title_entry.get():
            style_guide.show_error_in_entry(self.session_title_entry)


from typing import Tuple, Optional, Callable
from .data_type import IntegrationPoint,BasisType,Elements
from abc import ABC
from types import MappingProxyType

class ExpansionValidationException(Exception):
    """Custom exception for errors in expansion validation."""
    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

# TODO Note -> See MeshGraph.cpp/DefineBasisKeyFromExpansionTypeHomo
# ReadExpansionInfo etc. for more info in terms of how expansions are handled

class ExpansionDefinition:
    """Class responsible for defining a particular expansion depending on its shape
    """

    @staticmethod
    def _compute_tri_coefficients(num_modes_0: int,num_modes_1: int) -> int:
        """Helper function to compute number of unkown coefficients in a TRI element given the 
        number of modes in each direction.

        Args:
            num_modes_0 (int): Number of modes in direction 0
            num_modes_1 (int): Number of modes in direction 1

        Returns:
            int: Total number of unknown expansion coefficients for the defined TRI element
        """
        #definitions for these can be found in nektar/library/LibUtilities/BasicUtils/ShapeType.hpp
        return num_modes_0 * (num_modes_0 - 1) // 2 + num_modes_0 * (num_modes_1 - num_modes_0)

    @staticmethod
    def _compute_tet_coefficients(num_modes_0: int,num_modes_1: int,num_modes_2: int) -> int:
        """Helper function to compute number of unkown coefficients in a TET element given the 
        number of modes in each direction.

        Args:
            num_modes_0 (int): Number of modes in direction 0
            num_modes_1 (int): Number of modes in direction 1
            num_modes_2 (int): Number of modes in direction 2

        Returns:
            int: Total number of unknown expansion coefficients for the defined TET element
        """
        #definitions for these can be found in nektar/library/LibUtilities/BasicUtils/ShapeType.hpp
        num_coeffs = 0
        for a in range(num_modes_0):
            for b in range(num_modes_1 - a):
                for c in range(num_modes_2-a-b):
                    num_coeffs += 1

        return num_coeffs
    
    @staticmethod
    def _compute_pyr_coefficients(num_modes_0: int,num_modes_1: int,num_modes_2: int) -> int:
        """Helper function to compute number of unkown coefficients in a PYR element given the 
        number of modes in each direction.

        Args:
            num_modes_0 (int): Number of modes in direction 0
            num_modes_1 (int): Number of modes in direction 1
            num_modes_2 (int): Number of modes in direction 2

        Returns:
            int: Total number of unknown expansion coefficients for the defined PYR element
        """
        #definitions for these can be found in nektar/library/LibUtilities/BasicUtils/ShapeType.hpp
        num_coeffs = 0
        for a in range(num_modes_0):
            for b in range(num_modes_1):
                for c in range(num_modes_2-max(a,b)):
                    num_coeffs += 1

        return num_coeffs

    DIMENSIONS: MappingProxyType[Elements,int] = MappingProxyType({
                                        Elements.SEG: 1,
                                        Elements.QUAD: 2,Elements.TRI: 2,
                                        Elements.HEX: 3,Elements.TET: 3,Elements.PYR: 3, Elements.PRISM: 3
                                    })

    #definitions for these can be found in nektar/library/LibUtilities/BasicUtils/ShapeType.hpp
    NUMBER_OF_COEFFICIENTS: MappingProxyType[Elements,Callable[[Tuple[int,...]],int]] = MappingProxyType({
                                        Elements.SEG: lambda num_modes: num_modes[0],
                                        Elements.QUAD: lambda num_modes: num_modes[0] * num_modes[1],
                                        Elements.TRI: lambda num_modes: ExpansionDefinition._compute_tri_coefficients(num_modes[0],num_modes[1]),
                                        Elements.HEX: lambda num_modes: num_modes[0]*num_modes[1]*num_modes[2],
                                        Elements.PYR: lambda num_modes: ExpansionDefinition._compute_pyr_coefficients(num_modes[0],num_modes[1],num_modes[2]),
                                        Elements.PRISM: lambda num_modes: num_modes[1] * ExpansionDefinition._compute_tri_coefficients(num_modes[0],num_modes[2]),
                                        Elements.TET: lambda num_modes: ExpansionDefinition._compute_tet_coefficients(num_modes[0],num_modes[1],num_modes[2]),
                                    })

    def __init__(self,
                element: Elements,
                basis_type: Tuple[BasisType,...]=None,
                num_modes: Tuple[int,...]=None,
                integration_point_type: Tuple[IntegrationPoint,...]=None,
                num_points: Tuple[int,...]=None,
                fields: Tuple[str,...]=None) -> None:
        """Class initialiser

        Args:
            element (Elements): Element type
            basis_type (Tuple[BasisType,...], optional): Expansion basis in each principal direction (same size as element dimension). Defaults to None.
            num_modes (Tuple[int,...], optional): Number of expansion modes in each principal direction (same size as element dimension). Defaults to None.
            integration_point_type (Tuple[IntegrationPoint,...], optional): Integration point type in each principal direction (same size as element dimension). Defaults to None.
            num_points (Tuple[int,...], optional): Number of integration points in each principal direction (same size as element dimension). Defaults to None.
            fields (Tuple[str,...], optional): Fields associated with this expansion. Defaults to None.
        """

        self.element = element
        self.basis: Tuple[BasisType,...] = basis_type
        self.num_points: Tuple[int,...] = num_points
        self.integration_point_type: Tuple[IntegrationPoint,...] = integration_point_type
        self.num_modes: Tuple[int,...] = num_modes
        
        self.fields: Tuple[str] = fields
        
        #make sure only valid definitions lead to initialisation
        self._validate()

    def _validate(self):
        """Check expansion definition is correct. If fails validation, error will be thrown

        Raises:
            ExpansionValidationException: _description_
            ExpansionValidationException: _description_
            ExpansionValidationException: _description_
            ExpansionValidationException: _description_
            ExpansionValidationException: _description_
            ExpansionValidationException: _description_
            ExpansionValidationException: _description_
            ExpansionValidationException: _description_
        """
        expected_dim = ExpansionDefinition.DIMENSIONS[self.element]

        if len(self.num_modes) != expected_dim:
            raise ExpansionValidationException(f"Element {self.element} expects dimension {expected_dim}, modes are: {self.num_modes}")
        if len(self.num_points) != expected_dim:
            raise ExpansionValidationException(f"Element {self.element} expects dimension {expected_dim}, points are: {self.num_points}")
        if len(self.integration_point_type) != expected_dim:
            raise ExpansionValidationException(f"Element {self.element} expects dimension {expected_dim}, point types are: {self.integration_point_type}")
        if len(self.basis) != expected_dim:
            raise ExpansionValidationException(f"Element {self.element} expects dimension {expected_dim}, basis are: {self.basis}")

        #check based on rules in nektar/library/LibUtilies/BasicUtils/ShapeType.hpp 
        if self.element == Elements.TRI and self.num_modes[0] > self.num_modes[1]:
            raise ExpansionValidationException(f"Element {self.element} has invalid mode configuration: num_modes[0] ({self.num_modes[0]}) should not be greater than num_modes[1] ({self.num_modes[1]})")
        if self.element == Elements.TET and (self.num_modes[0] > self.num_modes[2] or self.num_modes[1] > self.num_modes[2]):
            raise ExpansionValidationException(f"Element {self.element} has invalid mode configuration: num_modes[0] ({self.num_modes[0]}) should not be greater than num_modes[1] ({self.num_modes[1]})")
        if self.element == Elements.PYR and (self.num_modes[0] > self.num_modes[2] or self.num_modes[1] > self.num_modes[2]):
            raise ExpansionValidationException(f"Element {self.element} has invalid mode configuration: num_modes[0] ({self.num_modes[0]}) should not be greater than num_modes[1] ({self.num_modes[1]})")
        if self.element == Elements.PRISM and self.num_modes[0] > self.num_modes[2]:
            raise ExpansionValidationException(f"Element {self.element} has invalid mode configuration: num_modes[0] ({self.num_modes[0]}) should not be greater than num_modes[1] ({self.num_modes[1]})")

    def get_num_coefficients(self):
        """Given the expansion definition, compute number of coefficients to be computed for each
        element of this expansion type.

        Returns:
            _type_: _description_
        """
        num_coeff_callable: Callable[[Tuple[int,...]],int] = ExpansionDefinition.NUMBER_OF_COEFFICIENTS[self.element]
        return num_coeff_callable(self.num_modes)

#preferred interface for constructing data
class ExpansionFactory(ABC):
    """Factory interface to define expected default expansion types. In Nektar++ session files, can define expansion with 
    some default construction, e.g. <E COMPOSITE="C[0]" NUMMODES="6" FIELDS="u,v" TYPE="MODIFIED" />. Here "MODIFIED" has 
    a default expansion construction as defined within `nektar/library/SpatialDomains/MeshGraph.cpp 
    <https://gitlab.nektar.info/nektar/nektar/-/blob/master/library/SpatialDomains/MeshGraph.cpp?ref_type=heads>`_.

    Raises:
        ExpansionValidationException: _description_
    """

    #only these four class variables should be overwritten in child classes for different behaviour
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]]
    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint,...]]
    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]]
    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]]

    def __init__(self):
        """Class initialiser
        """
        pass

    def get_expansion(self,element: Elements, nummodes: int,fields: Optional[Tuple[str,...]]=None) -> ExpansionDefinition:
        """Get the expansion definition from the factory

        Args:
            element (Elements): Element type to be created. Only shapes are allowed.
            nummodes (int): Number of modes in the expansion
            fields (Optional[Tuple[str,...]], optional): Fields in the expansion. Defaults to None.

        Raises:
            ExpansionValidationException: _description_

        Returns:
            ExpansionDefinition: Definition of the expansion
        """

        basis: Optional[BasisType] = self.BASIS_MAP.get(element,None)
        integration_points: Optional[IntegrationPoint] = self.INTEGRATION_POINTS_MAP.get(element,None)
        
        num_modes_callable: Optional[Callable[[int],Tuple[int,...]]] = self.NUM_MODES_MAP.get(element,None)
        num_points_callable: Optional[Callable[[int],Tuple[int,...]]] = self.NUM_POINTS_MAP.get(element,None)
        
        if basis and integration_points and num_modes_callable and num_points_callable:
            num_modes = num_modes_callable(nummodes)
            num_points = num_points_callable(nummodes)
            return ExpansionDefinition(element,basis,num_modes,integration_points,num_points,fields)
        else:
            raise ExpansionValidationException(f"{self.__class__.__name__} has no default definition for {element}")

#see nektar/library/SpatialDomains/MeshGraph.cpp for default expansions
class ModifiedExpansionFactory(ExpansionFactory):
    """Factory for creating MODIFIED default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
            Elements.SEG: (BasisType.MODIFIED_A,),
            Elements.TRI: (BasisType.MODIFIED_A, BasisType.MODIFIED_B),
            Elements.QUAD: (BasisType.MODIFIED_A, BasisType.MODIFIED_A),
            Elements.HEX: (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_A),
            Elements.PRISM: (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_B),
            Elements.PYR: (BasisType.MODIFIED_A, BasisType.MODIFIED_A, BasisType.MODIFIED_PYR_C),
            Elements.TET: (BasisType.MODIFIED_A, BasisType.MODIFIED_B, BasisType.MODIFIED_C),
        })
    
    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
            Elements.SEG: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,),
            Elements.QUAD: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
            Elements.TRI: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0),
            Elements.HEX: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
            Elements.PRISM: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0),
            Elements.PYR: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0),
            Elements.TET: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0),
        })
    
    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
            Elements.SEG: lambda nummodes: (nummodes,),
            Elements.QUAD: lambda nummodes: (nummodes, nummodes),
            Elements.TRI: lambda nummodes: (nummodes, nummodes),
            Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
            Elements.PRISM: lambda nummodes: (nummodes, nummodes, nummodes),
            Elements.PYR: lambda nummodes: (nummodes, nummodes, nummodes),
            Elements.TET: lambda nummodes: (nummodes, nummodes, nummodes),
        })

    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes,quad_offset=1: (nummodes+quad_offset,),
        Elements.QUAD: lambda nummodes,quad_offset=1: (nummodes+quad_offset,nummodes+quad_offset),
        Elements.TRI: lambda nummodes,quad_offset=1: (nummodes+quad_offset, nummodes+quad_offset-1),
        Elements.HEX: lambda nummodes, quad_offset=1: (nummodes+quad_offset, nummodes+quad_offset, nummodes+quad_offset),
        Elements.PRISM: lambda nummodes, quad_offset=1: (nummodes+quad_offset,nummodes+quad_offset,nummodes+quad_offset-1),
        Elements.PYR: lambda nummodes, quad_offset=1: (nummodes+quad_offset,nummodes+quad_offset,nummodes+quad_offset),
        Elements.TET: lambda nummodes, quad_offset=1: (nummodes+quad_offset, nummodes+quad_offset-1, nummodes+quad_offset-1),
    })

class ModifiedQuadPlus1ExpansionFactory(ModifiedExpansionFactory):
    """Factory for creating MODIFIED_QUAD_PLUS_1 default expansions
    """
    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes,quad_offset=2: (nummodes+quad_offset,),
        Elements.QUAD: lambda nummodes,quad_offset=2: (nummodes+quad_offset,nummodes+quad_offset),
        Elements.TRI: lambda nummodes,quad_offset=2: (nummodes+quad_offset, nummodes+quad_offset-1),
        Elements.HEX: lambda nummodes, quad_offset=2: (nummodes+quad_offset, nummodes+quad_offset, nummodes+quad_offset),
        Elements.PRISM: lambda nummodes, quad_offset=2: (nummodes+quad_offset,nummodes+quad_offset,nummodes+quad_offset-1),
        Elements.PYR: lambda nummodes, quad_offset=2: (nummodes+quad_offset,nummodes+quad_offset,nummodes+quad_offset),
        Elements.TET: lambda nummodes, quad_offset=2: (nummodes+quad_offset, nummodes+quad_offset-1, nummodes+quad_offset-1),
    })

class ModifiedQuadPlus2ExpansionFactory(ModifiedExpansionFactory):
    """Factory for creating MODIFIED_QUAD_PLUS_2 default expansions
    """
    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes,quad_offset=3: (nummodes+quad_offset,),
        Elements.QUAD: lambda nummodes,quad_offset=3: (nummodes+quad_offset,nummodes+quad_offset),
        Elements.TRI: lambda nummodes,quad_offset=3: (nummodes+quad_offset, nummodes+quad_offset-1),
        Elements.HEX: lambda nummodes, quad_offset=3: (nummodes+quad_offset, nummodes+quad_offset, nummodes+quad_offset),
        Elements.PRISM: lambda nummodes, quad_offset=3: (nummodes+quad_offset,nummodes+quad_offset,nummodes+quad_offset-1),
        Elements.PYR: lambda nummodes, quad_offset=3: (nummodes+quad_offset,nummodes+quad_offset,nummodes+quad_offset),
        Elements.TET: lambda nummodes, quad_offset=3: (nummodes+quad_offset, nummodes+quad_offset-1, nummodes+quad_offset-1),
    })

class ModifiedGLLRadau10ExpansionFactory(ModifiedExpansionFactory):
    """Factory for creating MODIFIED_GLL_RADAU_10 default expansions
    """
    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
            Elements.SEG: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,),
            Elements.QUAD: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
            Elements.TRI: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0),
            Elements.HEX: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
            Elements.PRISM: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0),
            Elements.PYR: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0),
            Elements.TET: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0),
        })

class GLLLagranageExpansionFactory(ExpansionFactory):
    """Factory for creating GLL_LAGRANGE default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
            Elements.SEG: (BasisType.GLL_LAGRANGE,),
            Elements.QUAD: (BasisType.GLL_LAGRANGE, BasisType.GLL_LAGRANGE),
            Elements.TRI: (BasisType.GLL_LAGRANGE, BasisType.ORTHO_B),
            Elements.HEX: (BasisType.GLL_LAGRANGE, BasisType.GLL_LAGRANGE, BasisType.GLL_LAGRANGE)
        })

    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
            Elements.SEG: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,),
            Elements.QUAD: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
            Elements.TRI: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0),
            Elements.HEX: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE)
        })
        
    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
            Elements.SEG: lambda nummodes: (nummodes,),
            Elements.QUAD: lambda nummodes: (nummodes, nummodes),
            Elements.TRI: lambda nummodes: (nummodes, nummodes),
            Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
        })

    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes,quad_offset=1: (nummodes+quad_offset,),
        Elements.QUAD: lambda nummodes,quad_offset=1: (nummodes+quad_offset,nummodes+quad_offset),
        Elements.TRI: lambda nummodes,quad_offset=1: (nummodes+quad_offset, nummodes+quad_offset-1),
        Elements.HEX: lambda nummodes, quad_offset=1: (nummodes+quad_offset, nummodes+quad_offset, nummodes+quad_offset),
    })
class GLLLagrangeSEMExpansionFactory(GLLLagranageExpansionFactory):
    """Factory for creating GLL_LAGRANGE_SEM default expansions
    """
    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes: (nummodes,),
        Elements.QUAD: lambda nummodes: (nummodes, nummodes),
        Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
    })

    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes: (nummodes,),
        Elements.QUAD: lambda nummodes: (nummodes,nummodes),
        Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
    })

class GaussLagrangeExpansionFactory(ExpansionFactory):
    """Factory for creating GAUSS_LAGRANGE default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
            Elements.SEG: (BasisType.GAUSS_LAGRANGE,),
            Elements.QUAD: (BasisType.GAUSS_LAGRANGE, BasisType.GAUSS_LAGRANGE),
            Elements.HEX: (BasisType.GAUSS_LAGRANGE, BasisType.GAUSS_LAGRANGE, BasisType.GAUSS_LAGRANGE)
        })

    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
            Elements.SEG: (IntegrationPoint.GAUSS_GAUSS_LEGENDRE,),
            Elements.QUAD: (IntegrationPoint.GAUSS_GAUSS_LEGENDRE, IntegrationPoint.GAUSS_GAUSS_LEGENDRE),
            Elements.HEX: (IntegrationPoint.GAUSS_GAUSS_LEGENDRE, IntegrationPoint.GAUSS_GAUSS_LEGENDRE, IntegrationPoint.GAUSS_GAUSS_LEGENDRE)
        })
        
    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
            Elements.SEG: lambda nummodes: (nummodes,),
            Elements.QUAD: lambda nummodes: (nummodes, nummodes),
            Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
        })

    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes: (nummodes,),
        Elements.QUAD: lambda nummodes: (nummodes,nummodes),
        Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
    })

class OrthogonalExpansionFactory(ExpansionFactory):
    """Factory for creating ORTHOGONAL default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
            Elements.SEG: (BasisType.ORTHO_A,),
            Elements.QUAD: (BasisType.ORTHO_A, BasisType.ORTHO_A),
            Elements.TRI: (BasisType.ORTHO_A, BasisType.ORTHO_B),
            Elements.TET: (BasisType.ORTHO_A, BasisType.ORTHO_B, BasisType.ORTHO_C)
        })

    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
            Elements.SEG: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE,),
            Elements.QUAD: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
            Elements.TRI: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0),
            Elements.TET: (IntegrationPoint.GAUSS_LOBATTO_LEGENDRE, IntegrationPoint.GAUSS_RADAU_M_ALPHA1_BETA0, IntegrationPoint.GAUSS_RADAU_M_ALPHA2_BETA0)
        })
        
    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
            Elements.SEG: lambda nummodes: (nummodes,),
            Elements.QUAD: lambda nummodes: (nummodes, nummodes),
            Elements.TRI: lambda nummodes: (nummodes, nummodes),
            Elements.TET: lambda nummodes: (nummodes, nummodes, nummodes),
        })

    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes: (nummodes+1,),
        Elements.QUAD: lambda nummodes: (nummodes+1,nummodes+1),
        Elements.TRI: lambda nummodes: (nummodes+1,nummodes),
        Elements.TET: lambda nummodes: (nummodes+1, nummodes, nummodes),
    })

class FourierExpansionFactory(ExpansionFactory):
    """Factory for creating FOURIER default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
            Elements.SEG: (BasisType.FOURIER,),
            Elements.QUAD: (BasisType.FOURIER, BasisType.FOURIER),
            Elements.HEX: (BasisType.FOURIER, BasisType.FOURIER, BasisType.FOURIER)
        })

    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
            Elements.SEG: (IntegrationPoint.FOURIER_EVENLY_SPACED,),
            Elements.QUAD: (IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.FOURIER_EVENLY_SPACED),
            Elements.HEX: (IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.FOURIER_EVENLY_SPACED)
        })
        
    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
            Elements.SEG: lambda nummodes: (nummodes,),
            Elements.QUAD: lambda nummodes: (nummodes, nummodes),
            Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
        })

    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes: (nummodes,),
        Elements.QUAD: lambda nummodes: (nummodes,nummodes),
        Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
    })

class FourierSingleModeExpansionFactory(FourierExpansionFactory):
    """Factory for creating FOURIER_SINGLE_MODE default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
        Elements.SEG: (BasisType.FOURIER_SINGLE_MODE,),
        Elements.QUAD: (BasisType.FOURIER_SINGLE_MODE, BasisType.FOURIER_SINGLE_MODE),
        Elements.HEX: (BasisType.FOURIER_SINGLE_MODE, BasisType.FOURIER_SINGLE_MODE, BasisType.FOURIER_SINGLE_MODE)
    })

    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
            Elements.SEG: (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED,),
            Elements.QUAD: (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED),
            Elements.HEX: (IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED, IntegrationPoint.FOURIER_SINGLE_MODE_SPACED)
        })

class FourierHalfModeReExpansionFactory(FourierSingleModeExpansionFactory):
    """Factory for creating FOURIER_HALF_MODE_RE default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
        Elements.SEG: (BasisType.FOURIER_HALF_MODE_RE,),
        Elements.QUAD: (BasisType.FOURIER_HALF_MODE_RE, BasisType.FOURIER_HALF_MODE_RE),
        Elements.HEX: (BasisType.FOURIER_HALF_MODE_RE, BasisType.FOURIER_HALF_MODE_RE, BasisType.FOURIER_HALF_MODE_RE)
    })



class FourierHalfModeImExpansionFactory(FourierSingleModeExpansionFactory):
    """Factory for creating FOURIER_HALF_MODE_IM default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
        Elements.SEG: (BasisType.FOURIER_HALF_MODE_IM,),
        Elements.QUAD: (BasisType.FOURIER_HALF_MODE_IM, BasisType.FOURIER_HALF_MODE_IM),
        Elements.HEX: (BasisType.FOURIER_HALF_MODE_IM, BasisType.FOURIER_HALF_MODE_IM, BasisType.FOURIER_HALF_MODE_IM)
    })

class ChebyshevExpansionFactory(ExpansionFactory):
    """Factory for creating CHEBYSHEV default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
        Elements.SEG: (BasisType.CHEBYSHEV,),
        Elements.QUAD: (BasisType.CHEBYSHEV, BasisType.CHEBYSHEV),
        Elements.HEX: (BasisType.CHEBYSHEV, BasisType.CHEBYSHEV, BasisType.CHEBYSHEV)
    })

    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
        Elements.SEG: (IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV,),
        Elements.QUAD: (IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV, IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV),
        Elements.HEX: (IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV, IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV, IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV)
    })

    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
            Elements.SEG: lambda nummodes: (nummodes,),
            Elements.QUAD: lambda nummodes: (nummodes, nummodes),
            Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
        })

    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.SEG: lambda nummodes: (nummodes,),
        Elements.QUAD: lambda nummodes: (nummodes,nummodes),
        Elements.HEX: lambda nummodes: (nummodes, nummodes, nummodes),
    })

class FourierChebyshevExpansionFactory(ExpansionFactory):
    """Factory for creating FOURIER_CHEBYSHEV default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
        Elements.QUAD: (BasisType.FOURIER, BasisType.CHEBYSHEV),
    })

    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
        Elements.QUAD: (IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV),
    })

    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
            Elements.QUAD: lambda nummodes: (nummodes, nummodes),
        })

    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.QUAD: lambda nummodes: (nummodes,nummodes),
    })

class ChebyshevFourierExpansionFactory(FourierChebyshevExpansionFactory):
    """Factory for creating CHEBYSHEV_FOURIER default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
        Elements.QUAD: (BasisType.CHEBYSHEV, BasisType.FOURIER),
    })

    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
        Elements.QUAD: (IntegrationPoint.GAUSS_GAUSS_CHEBYSHEV, IntegrationPoint.FOURIER_EVENLY_SPACED),
    })

class ModifiedFourierExpansionFactory(ExpansionFactory):
    """Factory for creating MODIFIED_FOURIER default expansions
    """
    BASIS_MAP: MappingProxyType[Elements,Tuple[BasisType,...]] = MappingProxyType({
        Elements.QUAD: (BasisType.FOURIER, BasisType.MODIFIED_A),
    })

    INTEGRATION_POINTS_MAP: MappingProxyType[Elements,Tuple[IntegrationPoint]] = MappingProxyType({
        Elements.QUAD: (IntegrationPoint.FOURIER_EVENLY_SPACED, IntegrationPoint.GAUSS_LOBATTO_LEGENDRE),
    })

    NUM_MODES_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
            Elements.QUAD: lambda nummodes: (nummodes, nummodes),
        })

    NUM_POINTS_MAP: MappingProxyType[Elements,Callable[[int],Tuple[int,...]]] = MappingProxyType({
        Elements.QUAD: lambda nummodes: (nummodes,nummodes+1),
    })


from typing import Dict,Any,List
from .user import InvenioUserInfo
from .relations import Relations
from .data_types import ResourceType,RESOURCE_TYPE_REVERSE_MAP

class InvenioMetadata:
    """Metadata model for handling metadata as prescribed in InvenioRDM. Uses a builder OOP style to 
    allow clients to enhance the metadata description.
    """
    def __init__(self,title: str,
                publication_date:str,
                creators: List[InvenioUserInfo],
                resource_type: ResourceType=ResourceType.DATASET):
        """Class Initialiser

        Args:
            title (str): Title of file uploads
            publication_date (str): Date of publication, in format YYYY-MM-DD
            creators (List[InvenioUserInfo]): List of authors
            resource_type (ResourceType, optional): Type of resource to be uploaded. Defaults to ResourceType.DATASET.
        """
        #mandatory fields
        self.title = title
        self.publication_date = publication_date
        self.creators: List[InvenioUserInfo] = creators
        self.resource_type = resource_type.value

        #optional fields
        self.version: str = None
        self.description: str = None
        self.publisher: str = None
        self.related_identifiers: List[Relations] = []

    def get_metadata_payload(self) -> Dict[str,Any]:
        """Get the metadata payload formatted in way InvenioRDM expects

        Returns:
            Dict[str,Any]: Metadata json payload
        """
        creators_payload: List[Dict[str,Any]] = []
        for creator in self.creators:
            creators_payload.append({
                "person_or_org": creator.get_info()
            })
        
        data = {
            "title": self.title,
            "publication_date": self.publication_date,
            "resource_type": {"id": self.resource_type},
            "creators": creators_payload
        }

        #only add to payload if they exist
        if self.version:
            data["version"] = self.version
        
        if self.description:
            data["description"] = self.description

        if self.publisher:
            data["publisher"] = self.publisher

        if self.related_identifiers:
            relation_list_json = [relation.to_json() for relation in self.related_identifiers]
            data["related_identifiers"] = relation_list_json

        return data

    def add_version(self, version: str) -> None:
        """Add the version number (of the software or dataset). Semantic versioning is recommended.

        Args:
            version (str): Version number
        """
        self.version = version

    def add_description(self, description: str) -> None:
        """Add a description of the record to be uploaded

        Args:
            description (str): Description
        """
        self.description = description

    def add_publisher(self, publisher: str="InvenioRDM") -> None:
        """Add publisher 

        Args:
            publisher (str, optional): Add publisher name. Defaults to "InvenioRDM".
        """
        self.publisher = publisher

    def add_related_identifier(self,relation: Relations):
        """Add related persistent identifiers

        Args:
            relation (Relations): A related identifier
        """
        self.related_identifiers.append(relation)

    def to_json_serialisable(self) -> Dict[str,Any]:        
        """Serialise the object as JSON

        Returns:
            Dict[str,Any]: Serialised json object
        """
        data = {
            "title": self.title,
            "publication_date": self.publication_date,
            "resource_type": self.resource_type,
            "creators": [creator.to_json_serialisable() for creator in self.creators]
        }

        if self.version:
            data["version"] = self.version
        
        if self.description:
            data["description"] = self.description

        if self.publisher:
            data["publisher"] = self.publisher

        if self.related_identifiers:
            relation_list_json = [relation.to_json() for relation in self.related_identifiers]
            data["related_identifiers"] = relation_list_json

        return data    
    
    @classmethod
    def from_json(cls,data: Dict[str,Any]) -> 'InvenioMetadata':
        """Reconstruct object from JSON serialisation

        Args:
            data (Dict[str,Any]): Serialised JSON data

        Returns:
            InvenioMetadata: Reconstructed object
        """
        title = data["title"]
        publication_date = data["publication_date"]
        resource_type: ResourceType = RESOURCE_TYPE_REVERSE_MAP[data["resource_type"]]
        creators: List[InvenioUserInfo] = [InvenioUserInfo.from_json(creator) for creator in data["creators"]]

        metadata = InvenioMetadata(title,publication_date,creators,resource_type)

        #only add the following optional data if present in serialisation
        if version := data.get("version",None):
            metadata.add_version(version)

        if description := data.get("description",None):
            metadata.add_description(description)

        if publisher := data.get("publisher",None):
            metadata.add_publisher(publisher)

        if related_identifiers := data.get("related_identifiers",None):
            for identifier_json in related_identifiers:
                import logging
                logging.error(identifier_json)
                metadata.add_related_identifier(Relations.from_json(identifier_json))

        return metadata


from .identifier import Identifier,IdentifierType
from .metadata import InvenioMetadata
from .user import InvenioPersonInfo,InvenioOrgInfo,InvenioUserInfo
from .data_types import RelationsSchemes,RelationType,ResourceType
from .relations import Relations
from .extractor import NekAutoExtractor

__all__ = [
    "Identifier", 
    "IdentifierType", 
    "InvenioMetadata", 
    "InvenioPersonInfo", 
    "InvenioOrgInfo", 
    "InvenioUserInfo", 
    "RelationsSchemes", 
    "RelationType", 
    "ResourceType", 
    "Relations", 
    "NekAutoExtractor"
]



from abc import ABC,abstractmethod
from typing import Dict,Any,List,Set,Union
from .identifier import Identifier,IdentifierType
class InvenioUserInfo(ABC):
    """Base abstract class describing information about a user
    """

    def __init__(self):
        self.id_schemes: Set[IdentifierType] = set()
        self.identifiers: List[Identifier] = []

    def add_identifier(self,identifier: Identifier) -> None:
        """Add a persistent identifier related to the user

        Args:
            identifier (Identifier): Object describing a unique identifier

        Raises:
            ValueError: _description_
        """
        id_type: IdentifierType = identifier.get_id_type()
        
        if id_type in self.id_schemes:
            raise ValueError(f"Cannot have duplicate identifiers of same type for one organisation {id_type}")        

        self.id_schemes.add(id_type)
        self.identifiers.append(identifier)

    @abstractmethod
    def get_info(self) -> Dict[str,Any]:
        """Return the json payload package

        Returns:
            Dict[str,Any]: JSON representation of the information, formatted for an API call
        """
        identifier_payload = []
        for identifier in self.identifiers:
            payload = {
                "scheme": identifier.get_id_type().value.lower(),
                "identifier": identifier.get_id()
            }
            identifier_payload.append(payload)

        data = {}
        if identifier_payload:
            data["identifiers"] = identifier_payload

        return data

    @abstractmethod
    def to_json_serialisable(self) -> Dict[str,Any]:
        """Return a json serialisable object representing this class

        Returns:
            Dict[str,Any]: JSON serialisable object
        """
        data = {}
        identifiers = [id.to_json_serialisable() for id in self.identifiers]
        if identifiers:
            data["identifiers"] = identifiers

        return data
    
    #TODO This currently works under the assumption that personal, organizational can be distinctly split
    # a better format with less assumptions could be done
    @classmethod
    @abstractmethod
    def from_json(cls,data: Dict[str,Any]) -> Union['InvenioOrgInfo','InvenioPersonInfo']:
        """Deserialise json serialisable object to reconstruct an object of this class. Factory method.

        Args:
            data (Dict[str,Any]): JSON serialisable object

        Returns:
            UserInfo: One of its sub-classes
        """
        type_map: Dict[str,InvenioUserInfo] = {
            "personal": InvenioPersonInfo,
            "organizational": InvenioOrgInfo
        }

        user_type = data.get("type")
        if user_type not in type_map:
            raise ValueError(f"Unknown user info type: {user_type}")

        return type_map[user_type].from_json(data)

    @abstractmethod
    def __eq__(self,other: object) -> bool:
        pass

class InvenioPersonInfo(InvenioUserInfo):
    """Holds information about a person in the format expected from InvenioRDM. Effectively a model for the 
    InvenioRDM Author as a Person field.
    """
    def __init__(self, given_name: str, family_name: str) -> None:
        """Class initialiser

        Args:
            given_name (str): First name
            family_name (str): Last name
        """
        super().__init__()

        self.type: str = "personal"
        self.given_name: str = given_name
        self.family_name: str = family_name

    def get_info(self) -> Dict[str,Any]:
        """Return the json payload package, formatted in a way InvenioRDM API expects

        Returns:
            Dict[str,Any]: json payload package, formatted in a way InvenioRDM API expects
        """        
        data = {
            "type": self.type,
            "given_name": self.given_name,
            "family_name": self.family_name,
        }

        return super().get_info() | data
    
    def __str__(self):
        return f"Person: {self.given_name} {self.family_name}"

    def to_json_serialisable(self) -> Dict[str,Any]:
        """Return a json serialisable object representing this class

        Returns:
            Dict[str,Any]: JSON serialisable object
        """
        data = {
            "type": self.type,
            "given_name": self.given_name,
            "family_name": self.family_name,
        }

        return super().to_json_serialisable() | data
    
    @classmethod
    def from_json(cls,data: Dict[str,Any]) -> 'InvenioPersonInfo':
        """Deserialise json serialisable object to reconstruct an object of this class.

        Args:
            data (Dict[str,Any]): JSON serialisable object

        Returns:
            InvenioPersonInfo: Reconstructed object
        """
        given_name = data["given_name"]
        family_name = data["family_name"]

        data_identifiers = data.get("identifiers",[])#in case not present in dict
        identifiers: List[Identifier] = [Identifier.from_json(id) for id in data_identifiers]
        
        person = InvenioPersonInfo(given_name,family_name)

        for identifier in identifiers:
            person.add_identifier(identifier)

        return person

    def __eq__(self,other: object) -> bool:
        if not isinstance(other, InvenioPersonInfo):
            return False
        
        return (
            self.type == other.type and
            self.given_name == other.given_name and
            self.family_name == other.family_name and
            self.identifiers == other.identifiers
        )

class InvenioOrgInfo(InvenioUserInfo):
    """Holds information about an organisation in the format expected from InvenioRDM. Effectively a model for the 
    InvenioRDM Author as an Organisation field.
    """
    def __init__(self, name: str) -> None:
        """Class initialiser

        Args:
            name (str): Organisation name
        """
        super().__init__()
        self.type: str = "organizational"
        self.name: str = name

    def get_info(self) -> Dict[str,Any]:
        """Return the json payload package, formatted in a way InvenioRDM API expects

        Returns:
            Dict[str,Any]: json payload package, formatted in a way InvenioRDM API expects
        """
        data = {
            "type": self.type,
            "name": self.name,
        }

        return super().get_info() | data

    def __str__(self):
        return f"Organisation: {self.name}"
    
    def to_json_serialisable(self):
        """Return a json serialisable object representing this class

        Returns:
            Dict[str,Any]: JSON serialisable object
        """
        data = {
            "type": self.type,
            "name": self.name,
        }

        return super().to_json_serialisable() | data
    
    @classmethod
    def from_json(cls,data: Dict[str,Any]) -> 'InvenioOrgInfo':
        """Deserialise json serialisable object to reconstruct an object of this class.

        Args:
            data (Dict[str,Any]): JSON serialisable object

        Returns:
            InvenioOrgInfo: Reconstructed object
        """
        name = data["name"]

        data_identifiers = data.get("identifiers",[])#in case not present in dict
        identifiers: List[Identifier] = [Identifier.from_json(id) for id in data_identifiers]
        
        org = InvenioOrgInfo(name)

        for identifier in identifiers:
            org.add_identifier(identifier)

        return org
    
    def __eq__(self,other: object) -> bool:
        if not isinstance(other, InvenioOrgInfo):
            return False
        
        return (
            self.type == other.type and
            self.name == other.name and
            self.identifiers == other.identifiers
        )


import logging
from typing import Optional, Type, List
import lxml.etree as ET
from .exceptions import XMLReaderException

class XMLReader:
    def __init__(self, filename: str, mode: str = "r") -> None:
        self.filename = filename
        self.mode = mode
        self.tree: Optional[ET.ElementTree] = None
    
    def __enter__(self) -> "XMLReader":
        try:
            self.tree = ET.parse(self.filename)
            return self # Return XMLReader instance
        except Exception as e:
            logging.fatal(f"Error opening XML file: {e}")
            raise e
        
    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[Type[BaseException]]) -> None:
        self.tree = None

    def get_structure(self) -> List[str]:
        """_summary_

        :raises XMLReaderException: _description_
        :return: _description_
        :rtype: List[str]
        """
        
        if not self.tree:
            raise XMLReaderException("XML tree is not loaded.")
        
        root = self.tree.getroot()
        structure = []

        def traverse(node, path=""):
            current_path = f"{path}/{node.tag}" if path else node.tag
            structure.append(current_path)
            for child in node:
                traverse(child, current_path)
        
        traverse(root)
        return structure
    
    def merge_first_level_elements_with(self, other: str, write_file: Optional[str] = None) -> None:
        """Merges two XML files. If first-level repeated elements are found, will REPLACE them."""
        if not self.tree:
            raise XMLReaderException("XML tree not loaded.")
        
        try:
            other_tree = ET.parse(other)
        except Exception as e:
            raise XMLReaderException(f"Error Parsing the XML file '{other}': {e}")
        
        root = self.tree.getroot()
        other_root = other_tree.getroot()

        # Store first-level elements of the original XML in a dictionary
        existing_elements = {child.tag: child for child in root}

        for child in other_root:
            if child.tag in existing_elements:
                # Replace existing element
                root.replace(existing_elements[child.tag], child)
            else:
                # Append new elements that don't exist
                root.append(child)

        if write_file:
            # **Step 1: Flatten the XML (remove unnecessary spaces and newlines)**
            xml_str = ET.tostring(root, encoding="UTF-8").decode()
            parser = ET.XMLParser(remove_blank_text=True)  # Strips extra spaces
            root = ET.XML(xml_str, parser)  # Reload into an XML object

            # **Step 2: Reapply indentation with 4 spaces**
            formatted_xml = ET.tostring(root, pretty_print=True, encoding="UTF-8").decode()
            formatted_xml = formatted_xml.replace("  ", "    ")  # Convert 2-space indent to 4-space indent

            # Write to file
            with open(write_file, "w", encoding="UTF-8") as f:
                f.write(formatted_xml)


import os
import h5py
from .hdf5_definitions import HDF5GroupDefinition,HDF5DatasetDefinition
from typing import Tuple,List,Optional
from .exceptions import OutputFileException,HDF5SchemaInconsistentException,HDF5SchemaExtraDefinitionException
from NekUpload.utils import parsing

class ValidateOutput:
    """Class responsible for all output file validation checks
    """
    def __init__(self, file_path: str):
        """Class initialiser

        Args:
            file_path (str): File path to output file
        """
        self.file = file_path
        self.file_name = os.path.basename(self.file)

    def check_schema(self) -> bool:
        """Check Output file conforms to HDF5 schema

        Raises:
            OutputSchemaHDF5Validator: _description_

        Returns:
            bool: Passed
        """
        try:
            with h5py.File(self.file, 'r') as f:
                self.schema_checker = OutputSchemaHDF5Validator(f)
                self.schema_checker.validate()
        except OSError as e:
            raise OutputFileException(self.file,f"Geometry file either does not exist or is not in HDF5 format {e}")

        return True
    
class OutputSchemaHDF5Validator:
    """Class for handling output HDF5 schema validation
    """

    NO_DIM_CONSTRAINTS = -1 #helper

    BASE_GROUPS = (HDF5GroupDefinition("NEKTAR",attributes=["FORMAT_VERSION"]),
                HDF5GroupDefinition("NEKTAR/Metadata",attributes=["ChkFileNum","Time"]), #this is bare minimum, depending on solver, can have more, also sessionFile#
                HDF5GroupDefinition("NEKTAR/Metadata/Provenance",attributes=["GitBranch","GitSHA1","Hostname","NektarVersion","Timestamp"]))

    EXPECTED_DATASETS = (HDF5DatasetDefinition("NEKTAR/DECOMPOSITION",(NO_DIM_CONSTRAINTS,)),)

    def __init__(self,f: h5py.File):
        """Class initialiser

        Args:
            f (h5py.File): Opened HDF5 file
        """
        self.file: h5py.File = f

    def validate(self):
        """Check whether specified file conforms to the HDF5 output schema
        """
        self._check_mandatory_groups(OutputSchemaHDF5Validator.BASE_GROUPS)
        self._check_mandatory_datasets(OutputSchemaHDF5Validator.EXPECTED_DATASETS)

        #acquire all other groups and datasets that should be present based on DECOMPOSITION definition
        self._assert_decomposition()
        expansion_groups: Tuple[HDF5GroupDefinition] = tuple(self._get_expansion_groups())
        optional_datasets: Tuple[HDF5DatasetDefinition] = tuple(self._get_optional_datasets())

        self._check_mandatory_groups(expansion_groups)
        self._check_mandatory_datasets(optional_datasets)
        
        #check no extraneous groups or datasets
        valid_groups: Tuple[HDF5GroupDefinition] = OutputSchemaHDF5Validator.BASE_GROUPS + expansion_groups
        valid_datasets: Tuple[HDF5DatasetDefinition] = OutputSchemaHDF5Validator.EXPECTED_DATASETS + optional_datasets

        valid_groups_str = [group.get_path() for group in valid_groups]
        valid_datasets_str = [dataset.get_path() for dataset in valid_datasets]
        self._check_only_valid_groups_exist(valid_groups_str)
        self._check_only_valid_datasets_exist(valid_datasets_str)

        #check some more DECOMPOSITION data


    def _check_mandatory_groups(self,groups: Tuple[HDF5GroupDefinition]):
        """Check whether mandatory HDF5 Groups are present in the file

        Args:
            groups (Tuple[HDF5GroupDefinition]): List of mandatory HDF5 Group definitions
        """
        for group in groups:
            group.validate(self.file)

    def _check_mandatory_datasets(self,datasets: Tuple[HDF5DatasetDefinition]):
        """CHeck whether mandatory HDF5 Datasets are present in the file

        Args:
            datasets (Tuple[HDF5DatasetDefinition]): List of mandatory HDF5 Dataset definitions
        """
        for dataset in datasets:
            dataset.validate(self.file)

    def _assert_decomposition(self):
        """Assert decomposition has correct shape

        Raises:
            HDF5SchemaInconsistentException: _description_
        """
        #decomposition should come in group of 7
        if self.file["NEKTAR/DECOMPOSITION"].shape[0] % 7 != 0:
            raise HDF5SchemaInconsistentException(self.file,"HDF5 Schema Error: Decomposition shape should be multiple of 7")

    def _get_expansion_groups(self) -> List[HDF5GroupDefinition]:
        """Get the expansion groups that should be defined, based on what is in DECOMPOSITION

        Raises:
            HDF5SchemaInconsistentException: _description_

        Returns:
            List[HDF5GroupDefinition]: _description_
        """
        decomposition_dataset: h5py.Dataset = self.file["NEKTAR/DECOMPOSITION"]
        #last of the 7 is a hash pointing to location in HDF5 file containing expansion data
        num_expansion_groups = decomposition_dataset.shape[0] // 7

        expected_groups: List[HDF5GroupDefinition] = []
        for i in range(6,7*num_expansion_groups,7):
            hash = decomposition_dataset[i]
            expected_groups.append(HDF5GroupDefinition(f"NEKTAR/{hash}",attributes=["BASIS","FIELDS","NUMMODESPERDIR","SHAPE"]))

        return expected_groups
    
    def _get_optional_datasets(self) -> List[HDF5DatasetDefinition]:
        """Get all optional datasets defined by DECOMPOSITION

        Returns:
            List[HDF5DatasetDefinition]: _description_
        """
        optional_datasets: List[HDF5DatasetDefinition] = []

        optionals = {"NEKTAR/ELEMENTIDS": 0,
                    "NEKTAR/DATA": 1,
                    "NEKTAR/POLYORDERS": 2,
                    "NEKTAR/HOMOGENEOUSYIDS": 3,
                    "NEKTAR/HOMOGENEOUSZIDS": 4,
                    "NEKTAR/HOMOGENEOUSSIDS": 5}

        for name,idx in optionals.items():
            if dataset := self._get_dataset_defined_in_decomposition(name,idx):
                optional_datasets.append(dataset)

        return optional_datasets
    
    def _get_dataset_defined_in_decomposition(self,
                                            dataset_name: str,
                                            decomposition_entry_id: int) -> Optional[HDF5DatasetDefinition]:
        """DECOMPOSITION contains sequence of 7 entries, some of which will lead to definition of
        extra datasets within the file. When the following are non-zero, a dataset is expected, and
        are constructed with the same rule:

        Note starting from 0:
        2 -> number of modes when variable polynomial is defined
        3 -> number of y planes for homogeneous simulations
        4 -> number of z planes for homogeneous simulations
        5 -> number of strips for homogeneous simulations

        Args:
            dataset_name (str): Name of the dataset to be defined
            decomposition_entry_id (int): Decomposition entry id for desired dataset 

        Returns:
            Optional[HDF5DatasetDefinition]: Dataset schema definition if one is required
        """
        decomposition_dataset: h5py.Dataset = self.file["NEKTAR/DECOMPOSITION"]
        size = decomposition_dataset.shape[0]
        num_data_points: int = 0

        for i in range(decomposition_entry_id,size,7):
            num_data_points += decomposition_dataset[i]

        return HDF5DatasetDefinition(dataset_name,(num_data_points,)) if num_data_points > 0 else None

    def _get_polyorder_dataset(self) -> Optional[HDF5DatasetDefinition]:
        """Get the polyorder dataset definition if it should exist, based on DECOMPOSITION entries, every third entry

        Returns:
            Optional[HDF5DatasetDefinition]: If polyorder dataset is defined, definition is returned, else None
        """
        decomposition_dataset: h5py.Dataset = self.file["NEKTAR/DECOMPOSITION"]
        size = decomposition_dataset.shape[0]
        #3rd of the 7 grouping in decomposition
        #is a number of modes that are polyorder???
        num_polyorder_modes: int = 0

        for i in range(2,size,7):
            num_polyorder_modes += decomposition_dataset[i]

        return HDF5DatasetDefinition("NEKTAR/POLYORDERS",(num_polyorder_modes,)) if num_polyorder_modes > 0 else None

    def _check_only_valid_groups_exist(self,valid_groups: List[str]):
        """Check that only valid groups exist.

        Args:
            valid_groups (str): List of paths for valid HDF5 Groups
        """
        #plus one to search for any extra invalid groups
        #"" is a valid group too, and is provided in function call
        valid_groups.append("")
        max_groups = len(valid_groups) + 1 
        groups = parsing.get_hdf5_groups_with_depth_limit(self.file,3,max_groups=max_groups)

        for group in groups:
            if group not in valid_groups:
                raise HDF5SchemaExtraDefinitionException(self.file,f"Encountered unkown group: {group}")

    def _check_only_valid_datasets_exist(self,valid_datasets: List[str]):
        """Check that only valid datasets exist.

        Args:
            valid_datasets (str): List of paths for valid HDF5 Datasets
        """
        max_datasets = len(valid_datasets) + 1
        datasets = parsing.get_hdf5_datasets_with_depth_limit(self.file,3,max_datasets=max_datasets)
        for dataset in datasets:
            if dataset not in valid_datasets:
                raise HDF5SchemaExtraDefinitionException(self.file,f"Encountered unkown dataset: {dataset}")



from abc import ABC, abstractmethod
import h5py
from typing import List,Tuple
from .exceptions import HDF5SchemaExistenceException,HDF5SchemaInconsistentException
from dataclasses import dataclass,field

class HDF5Definition(ABC):
    @abstractmethod
    def validate(self,h5py_file: h5py.File):
        pass

@dataclass(frozen=True)
class HDF5GroupDefinition(HDF5Definition):
    """Given an HDF5 file, responsible for checking group conforms to correct structure and contains
    the specified attributes. This is not an exclusive check, other non-specified attributes can also be present.
    All exceptions raised from this class are of type HDF5SchemaException or its children.
    """
    path: str
    attributes: List[str] = field(default_factory=list)

    def validate(self,f: h5py.File) -> bool:
        """Checks whether the defined HDF5 group is present and well-defined in the specified HDF5 file

        Args:
            f (h5py.File): HDF5 file

        Raises:
            HDF5SchemaException: _description_
            HDF5SchemaException: _description_
            HDF5SchemaException: _description_

        Returns:
            bool: _description_
        """
        if self.path not in f:
            raise HDF5SchemaExistenceException(f,f"HDF5 schema error, {self.path} is not in file")
        
        group: h5py.Group = f[self.path]
        if not isinstance(group, h5py.Group): 
            raise HDF5SchemaInconsistentException(f,f"HDF5 schema error, {self.path} is a {type(group)},not a group")

        #check attributes exist
        if not set(self.attributes).issubset(group.attrs.keys()):
            missing_attributes = set(self.attributes) - set(group.attrs.keys())
            if missing_attributes:
                raise HDF5SchemaInconsistentException(f, f"HDF5 schema error, missing attributes in {self.path}: {missing_attributes}")
    
        return True

    def __str__(self):
        return self.path

    def get_path(self):
        return self.path

@dataclass(frozen=True)
class HDF5DatasetDefinition(HDF5Definition):
    """Given an HDF5 file, responsible for checking if dataset conforms to schema expectations, such as shape constraints.
        All exceptions raised from this class are of type HDF5SchemaException or its children.
    """
    path: str
    dataset_shape: Tuple[int,...] = ()

    def validate(self, f: h5py.File) -> bool:
        """Checks whether the defined HDF5 dataset is present and well-defined in the specified HDF5 file

        Args:
            f (h5py.File): _description_

        Raises:
            HDF5SchemaException: _description_
            HDF5SchemaException: _description_
            HDF5SchemaException: _description_
            HDF5SchemaException: _description_

        Returns:
            bool: _description_
        """
        if self.path not in f:
            raise HDF5SchemaExistenceException(f,f"HDF5 schema error, {self.path} is not in file")
        
        dataset: h5py.Dataset = f[self.path]
        if not isinstance(dataset,h5py.Dataset):
            raise HDF5SchemaInconsistentException(f,f"HDF5 schema error, {self.path} is a {type(dataset)}, not a dataset")
        
        #check dataset shape if it exists
        if self.dataset_shape:
            actual_shape: Tuple[int,...] = dataset.shape
            
            #check expected dimension
            if len(actual_shape) != len(self.dataset_shape):
                raise HDF5SchemaInconsistentException(f,f"HDF5 schema error, {self.path} has dataset shape {actual_shape}, but expecting {self.dataset_shape}")
            
            for size,constrained_size in zip(actual_shape,self.dataset_shape):
                # negatives denotes no size restriction on dataset shape
                #so only positive ones are constraints
                if constrained_size >= 0 and constrained_size != size:
                    raise HDF5SchemaInconsistentException(f,f"HDF5 schema error, {self.path} has dataset shape {actual_shape}, but expecting {self.dataset_shape}")
                
        return True
    
    def __str__(self):
        return self.path

    def get_path(self):
        return self.path


import tkinter as tk
import ttkbootstrap as ttk
from ttkbootstrap.constants import *

class ScrolledListbox(ttk.Frame):
    def __init__(self,parent):
        super().__init__(
            master=parent
        )

        self.rowconfigure(0,weight=1)
        self.columnconfigure(0,weight=1)

        self._listbox = tk.Listbox(self,selectmode=EXTENDED)
        scrollbar_y= ttk.Scrollbar(self,command=self._listbox.yview)
        scrollbar_x = ttk.Scrollbar(self,command=self._listbox.xview,orient=HORIZONTAL)
        self._listbox.config(yscrollcommand=scrollbar_y.set,xscrollcommand=scrollbar_x)
        scrollbar_y.config(command=self._listbox.yview)
        scrollbar_x.config(command=self._listbox.xview)

        self._listbox.grid(row=0,column=0,sticky=(NSEW))
        scrollbar_x.grid(row=1,column=0,sticky=(W,E))
        scrollbar_y.grid(row=0,column=1,sticky=(N,S,W))

    @property
    def listbox(self):
        return self._listbox


import logging
from typing import Dict,Any,List
import requests
from .exceptions import APIError,ClientError
import os

"""
This file contains python wrapping of API calls for ease of use. 
Not using class here as each method should be atomic and stateless, much like REST API.
It is not the responsibility of this module to validate input, that is the client's job.
All inputs stated in the documentation should be provided.
Always return a response, so clients have flexibility to choose how they want to process things.
For now, lets not worry about other error codes
"""

# TODO Currently metadata expects {"metadata": metadata_payload},
# would be better for user to just pass metadata_payload, then construct above inside class
def create_draft_record(url: str, 
                        token: str,
                        metadata: Dict[str,Any],
                        custom_fields:Dict[str,Any] = None,
                        upload_file_enabled: bool=True) -> requests.Response:
    """Create a recrod draft in an InvenioRDM instance with the specified metadata.

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        metadata (Dict[str,Any], optional): Metadata to be uploaded.
        custom_fields (Dict[str,Any], optional): Custom fields for record (InvenioRDM v10 and newer). Defaults to None.
        upload_file_enabled (bool, optional): Denotes whether file upload allowed or not. Defaults to True.

    Raises:
        APIError: If an error occurs during the API call. 

    Returns:
        requests.Response: Returns a request.Response obhject on success (201 status code)
    """

    records_url = url.rstrip("/") + "/api/records"
    
    header = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {token}"
    }

    #add other payloads to metadata, create new copy to avoid mutability issues
    payload = {**metadata, "files": {"enabled": upload_file_enabled}}

    if custom_fields:
        payload["custom_fields"] = custom_fields

    try:
        response = requests.post(records_url, headers=header, json=payload)
        response.raise_for_status() # Raise an exception for bad status codes (4xx or 5xx)
        
        if response.status_code == 201:
            _log_debug_response("Record created succesfully",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response) 
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

# TODO Currently metadata expects {"metadata": metadata_payload},
# would be better for user to just pass metadata_payload, then construct above inside class
def update_draft_record(url: str, 
                        token: str,
                        record_id: str,
                        metadata: Dict[str,Any],
                        custom_fields:Dict[str,Any] = None,
                        upload_file_enabled: bool=True) -> requests.Response:
    """Update a record draft in InvenioRDM with the specified metadata

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        metadata (Dict[str,Any], optional): Metadata to be uploaded.
        custom_fields (Dict[str,Any], optional): Custom fields for record (InvenioRDM v10 and newer). Defaults to None.
        upload_file_enabled (bool, optional): Denotes whether file upload allowed or not. Defaults to True.

    Raises:
        APIError: If an error occurs during the API call. 

    Returns:
        requests.Response: Returns a request.Response obhject on success (200 status code)
    """

    records_url = url.rstrip("/") + f"/api/records/{record_id}/draft"
    
    header = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {token}"
    }

    #add other payloads to metadata, create new copy to avoid mutability issues
    payload = {**metadata, "files": {"enabled": upload_file_enabled}}

    if custom_fields:
        payload["custom_fields"] = custom_fields

    try:
        response = requests.put(records_url, headers=header, json=payload)
        response.raise_for_status()
        
        if response.status_code == 200:
            _log_debug_response("Record updated succesfully",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response) 
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def create_new_record_version(url:str,token:str,record_id: str) -> requests.Response:
    """Create a new draft record from a published record, thereby creating a new version

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Record ID from which new record version will be created

    Raises:
        APIError: If an error occurs during the API call. 

    Returns:
        requests.Response: Returns a request.Response obhject on success (201 status code)
    """
    record_url = url.rstrip("/") + f"/api/records/{record_id}/versions"
    header = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {token}"
    }

    try:
        response = requests.post(record_url,headers=header)
        response.raise_for_status()

        if response.status_code == 201:
            _log_debug_response("New record version created", response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def reserve_doi_draft(url:str,token:str,record_id: str) -> requests.Response:
    """Reserve a DOI for the specified draft record

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Record ID for which DOI reservation is for

    Raises:
        APIError: If an error occurs during the API call. 

    Returns:
        requests.Response: Returns a request.Response obhject on success (201 status code)
    """

    api_url = f"{url.rstrip('/')}/api/records/{record_id}/draft/pids/doi"
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
    }

    try:
        response = requests.post(api_url,headers=headers)
        response.raise_for_status()

        if response.status_code == 201:
            _log_debug_response("Reserved DOI", response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def prepare_file_upload(url: str,token: str,record_id: str,file_name_list: List[str]) -> requests.Response:
    """Creates a location in the Invenio database record to store the files. Capable of batch file preparation.

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Record ID for which files will be uploaded to
        file_name_list (List[str]): List of names of file to be uploaded

    Returns:
        requests.Response: Response for preparing file upload (201 response code on success)

    Raises:
        APIError: If an error occurs during the API call. 
        ClientError: If an error occurs due to invalid message parameters
    """

    header = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {token}"
    }

    #convert to JSON array with []
    body = [{"key": file_name} for file_name in file_name_list]
    
    record_url = url.rstrip("/") + f"/api/records/{record_id}/draft/files"

    try:
        response = requests.post(record_url,headers=header, json=body)
        response.raise_for_status()

        if response.status_code == 201:
            _log_debug_response("Files prepared for uploads", response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            logging.error(err_msg)
            raise APIError(err_msg,response=response)
    except requests.exceptions.RequestException as e:
        err_msg = f"Request Error: {e}"
        logging.error(err_msg)
        raise APIError(err_msg)

def upload_file(url: str, token: str, record_id: str, file_path: str) -> requests.Response:
    """Upload the specified file to the specified draft record

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Record ID for which files will be uploaded to
        file_path (str): Path of file to be uploaded

    Returns:
        requests.Response: Upload response object
    
    Raises:
        APIError: If an error occurs during the API call. 
    """

    #there is an optional size parameter, but we choose to ignore for now
    header = {
        "Content-Type": "application/octet-stream",
        "Authorization": f"Bearer {token}"
    }

    filename = file_path.split('/')[-1]
    file_upload_url = url.rstrip("/") + f"/api/records/{record_id}/draft/files/{filename}/content"

    #open in binary mode, requests behaviour is file is streamed for upload, avoiding memory issues
    with open(file_path, "rb") as f:
        try:
            response = requests.put(file_upload_url, headers=header, data=f)
            response.raise_for_status()
            
            if response.status_code == 200:
                _log_debug_response(f"File {file_path} uploaded successfully", response)
                return response
            else:
                err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
                raise APIError(err_msg,response=response)
        except requests.exceptions.RequestException as e:
            raise APIError(f"Request Error: {e}")

def commit_file_upload(url: str, token: str, record_id: str, filename: str) -> requests.Response:
    """Save uploaded file in the draft record

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Record ID for which files are uploaded to
        filename (str): Name of file to be uploaded

    Returns:
        requests.Response: Commit response object

    Raises:
        APIError: If an error occurs during the API call. 
    """

    header = {"Authorization": f"Bearer {token}"}    
    file_commit_url = url.rstrip("/") + f"/api/records/{record_id}/draft/files/{filename}/commit"

    try:
        response = requests.post(file_commit_url, headers=header)
        response.raise_for_status()
        
        if response.status_code == 200:
            _log_debug_response(f"File {filename} committed successfully",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)    
    except requests.exceptions.RequestException as e:
            raise APIError(f"Request Error: {e}")

def publish_draft(url: str, token: str, record_id: str) -> requests.Response:
    """Publish the specified draft on InvenioRDM

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Id of record to be published

    Returns:
        requests.Response: Publish response object

    Raises:
        APIError: If an error occurs during the API call. 
        ClientError: If an error occurs due to invalid message parameters
    """
        
    header = {"Authorization": f"Bearer {token}"}
    publish_url = url.rstrip("/") + f"/api/records/{record_id}/draft/actions/publish"

    try:
        response = requests.post(publish_url, headers=header)
        response.raise_for_status() #raise exception for bad status codes
        
        if response.status_code == 202:
            _log_debug_response(f"Draft published successfully.",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)     
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def delete_draft(url: str, token: str, record_id: str) -> requests.Response:
    """Delete a draft record

    Args:
        url (str): Base url route to the invenio database, of form http:// or https://
        token (str): Personal access token
        record_id(str): Id of record to be prepared for file upload

    Returns:
        requests.Response: Delete response object

    Raises:
        APIError: If an error occurs during the API call. 
    """

    delete_url = url.rstrip("/") + f"/api/records/{record_id}/draft"

    header = {"Authorization": f"Bearer {token}"}

    try:
        response = requests.delete(delete_url, headers=header)
        response.raise_for_status() #raise exception for bad status codes

        if response.status_code == 204:
            _log_debug_response(f"Draft deleted successfully.",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)     
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def get_community(url:str, token: str,community_slug: str) -> requests.Response:
    """Get community specified by the community slug or id

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        community_slug (str): Community url slug or uuid

    Returns:
        requests.Response: Get community response

    Raises:
        APIError: If an error occurs during the API call. 
    """
    header = {"Authorization": f"Bearer {token}"}

    community_url = url.rstrip("/") + f"/api/communities/{community_slug}"

    try:
        response = requests.get(community_url,headers=header)
        response.raise_for_status()

        if response.status_code == 200:
            _log_debug_response(f"Successfully found community associated with {url}",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)     
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def submit_record_to_community(url: str,token: str,community_uuid:str, record_id:str) -> requests.Request:
    """Submit a specified record to a specified community

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        community_uuid (str): Community UUID
        record_id (str): Id of record to be submitted to community

    Returns:
        requests.Request: Community submission response
    
    Raises:
        APIError: If an error occurs during the API call. 
        ClientError: If an error occurs due to invalid message parameters
    """
    
    community_url = url.rstrip("/") + f"/api/records/{record_id}/draft/review"

    header = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {token}"
        }

    body =  {
        "receiver" : {
            "community": f"{community_uuid}"
        },
        "type": "community-submission"
    }

    try:
        response = requests.put(community_url,headers=header,json=body)
        response.raise_for_status()

        if response.status_code == 200:
            _log_debug_response(f"Successfully submitted record {record_id} to community {community_uuid}",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)     
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def submit_record_for_review(url: str,token: str,record_id: str,payload: Dict[str,str]) -> requests.Response:
    """Once record is submmitted to community, submit it for review to community admins

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Id of record to be submitted for review
        payload(str): Contains content and format
        
    Returns:
        requests.Response: Record review submission response
    
    Raises:
        APIError: If an error occurs during the API call. 
    """    
    submit_review_url = url.strip("/") + f"/api/records/{record_id}/draft/actions/submit-review"

    header = {"Authorization": f"Bearer {token}"}
    body = {"payload": payload}

    try: 
        response = requests.post(submit_review_url,headers=header,json=body)
        response.raise_for_status()

        if response.status_code == 202:
            _log_debug_response(f"Record submitted to community for review",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)     
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def get_draft_record(url: str, token: str, record_id: str) -> requests.Response:
    """Get draft record associated with record_id

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Id of desired record
        
    Returns:
        requests.Response: Record review submission response
    
    Raises:
        APIError: If an error occurs during the API call. 
    """
    draft_url = url.rstrip("/") + f"/api/records/{record_id}/draft"

    header = {"Authorization": f"Bearer {token}"}
    
    try: 
        response = requests.get(draft_url,headers=header)
        response.raise_for_status()

        if response.status_code == 200:
            _log_debug_response(f"Record {record_id} acquired",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)     
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def delete_review_request(url:str,token:str,record_id:str) -> requests.Response:
    """Delete review request for the specified draft

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Id of record foe which review request will be deleted
    
    Returns:
        requests.Response: Record review submission response
    
    Raises:
        APIError: If an error occurs during the API call. 
        ClientError: If an error occurs due to invalid message parameters
    """
    review_url = url.rstrip("/") + f"/api/records/{record_id}/draft/review"

    header = {"Authorization": f"Bearer {token}"}

    try: 
        response = requests.delete(review_url,headers=header)
        response.raise_for_status()

        if response.status_code == 204:
            _log_debug_response(f"Record {record_id} acquired",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)     
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")   

def cancel_review_request(url:str,token:str,request_id:str,payload: Dict[str,str]) -> requests.Response:
    """Cancel a user-submitted review request. Only request's creator can cancel it

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Id of record for which review request will be cancelled
        payload(str): Contains content and format

    Returns:
        requests.Response: Record review submission response
    
    Raises:
        APIError: If an error occurs during the API call. 
    """    
    cancel_request = url.rstrip("/") + f"/api/requests/{request_id}/actions/cancel"

    header = {"Authorization": f"Bearer {token}"}
    body = {"payload": payload}

    try: 
        response = requests.post(cancel_request,headers=header,json=body)
        response.raise_for_status()

        if response.status_code == 200:
            _log_debug_response(f"Request {request_id} cancelled.",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)     
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")   
    
def get_record(url:str,token:str,record_id:str) -> requests.Response:
    """Get published record associated with record id

    Args:
        url (str): Base url route to the invenio database
        token (str): Personal access token
        record_id (str): Record ID of desired record
    
    Raises:
        APIError: If API call fails

    Returns:
        requests.Response: _description_
    """
    header = {"Authorization": f"Bearer {token}"}
    get_request = url.rstrip("/") + f"/api/records/{record_id}"

    try: 
        response = requests.get(get_request,headers=header)
        response.raise_for_status()

        if response.status_code == 200:
            _log_debug_response(f"Record {record_id} retrieved.",response)
            return response
        else:
            err_msg = f"Unexpected status code: {response.status_code} - {response.text}"
            raise APIError(err_msg,response=response)     
    except requests.exceptions.RequestException as e:
        raise APIError(f"Request Error: {e}")

def _log_debug_response(msg: str, response: requests.Response) -> None:
    """Log a debug statement to logger, with message and response.
    Will take form msg: response. For long responses, they are shortened when logged. 

    Args:
        msg (str): User defined message
        response (requests.Response): Request response 
    """
    try:
        response_data = response.json()
        if isinstance(response_data, dict) and len(response_data) > 10:
            response_data = {k: response_data[k] for k in list(response_data)[:10]}
            response_data['...'] = '...'
        elif isinstance(response_data, list) and len(response_data) > 10:
            response_data = response_data[:10] + ['...']
    except ValueError:
        response_data = response.text[:1000] + '...' if len(response.text) > 1000 else response.text

    logging.debug(f"{msg}: {response_data}")


from lxml import etree
from .exceptions import XMLSchemaException,MissingInputFileException,MissingOutputFileException
import os
from typing import List,Dict
from NekUpload.utils import parsing

class ValidateSession:
    """Class responsible for validation session files
    """
    def __init__(self,file_path: str):
        """Class initialiser

        Args:
            file_path (str): File path to the session XML file
        """
        self.file_path = file_path
        self.xml_tree = self._load_DOM_tree(self.file_path)

    def _load_DOM_tree(self, xml_file: str) -> etree._Element:
        """Loads the DOM tree of the XML file

        Args:
            xml_file (str): Path to XML file

        Returns:
            etree._Element: Internal representation of DOM tree
        """
        with open(xml_file, "rb") as xml:
            xml_tree = etree.XML(xml.read())
        
        return xml_tree

    def is_valid_xml(self,xml_file: str,schema_file_path: str) -> bool:
        """Checks whether XML file conforms to a schema

        Args:
            xml_file (str): XML file path
            schema_file_path (str): XSD schema file path

        Raises:
            XMLSchemaException: _description_

        Returns:
            bool: Passed
        """
        xsd_file = schema_file_path
        
        with open(xsd_file,"rb") as xsd:
            schema_root = etree.XML(xsd.read())
            schema = etree.XMLSchema(schema_root)

        with open(xml_file,"rb") as xml:
            xml_tree = etree.XML(xml.read())
        
        if schema.validate(xml_tree):
            return True
        else:
            raise XMLSchemaException(self.file_path,schema.error_log)
    
    def check_schema(self) -> bool:
        """Check file conforms to XML session schema

        Returns:
            bool: Passed
        """
        xsd_schema = os.path.join(os.path.dirname(__file__), 'schemas/nektar.xsd') #ensure path always correct
        return self.is_valid_xml(self.file_path, xsd_schema)
        
    def check_file_dependencies(self, files: List[str]) -> bool:
        """Check all files associated with this session file are present

        Args:
            files (List[str]): Other files

        Raises:
            MissingInputFileException: _description_
            MissingOutputFileException: _description_

        Returns:
            bool: Passed
        """
        #check geometry files exist
        geometry: etree._Element = self.xml_tree.find("GEOMETRY")
        expected_geometry_file = geometry.attrib.get("HDF5FILE")

        if expected_geometry_file not in files:
            raise MissingInputFileException(expected_geometry_file,"Geometry file is missing")
        
        #check correct number of checkpoint files
        conditions: etree._Element = self.xml_tree.find("CONDITIONS")
        params: etree._Element = conditions.find("PARAMETERS")

        #params has a bunch of p child elements, each with different 
        p_dict: Dict[str,str] = {}
        for p in params:
            content = p.text #containss an equals
            param_name,value = parsing.get_both_sides_of_equals(content)
            p_dict[param_name] = value

        #there is a possibility that it references other params, so store all in dict
        p_dict = parsing.evaluate_parameters(p_dict)

        num_steps = p_dict["NumSteps"]
        chk_steps = p_dict["IO_CheckSteps"]

        #at step=0, checkpoint file 0 is generated, hence +1
        num_chk_files = num_steps // chk_steps + 1

        chk_files = parsing.get_all_files_with_extension(files,".chk")

        if num_chk_files != len(chk_files):
            raise MissingOutputFileException(chk_files,
                                            f"There are {len(chk_files)}. Should have {num_chk_files}")

        return True


import ttkbootstrap as ttk
from ttkbootstrap.constants import *
import tkinter as tk
from typing import Callable
from ttkbootstrap.window import Toplevel
from NekUpload.frontend import style_guide

class CreateAuthorPersonWindow(Toplevel):
    """Window containing form to create author (person)

    Args:
        Toplevel (_type_): tkinter Window object
    """
    def __init__(self,root, submit_function: Callable):
        """
        Initializes the CreateAuthorWindow.
        Args:
            root (Tk): The root window of the Tkinter application.
            submit_function (Callable): The function to be called when the submit button is pressed.
        """
        super().__init__(root)

        self.title("Specify Author Information") 
        self.grid_rowconfigure(0,weight=1)
        self.grid_columnconfigure(0,weight=1)

        content_frame = ttk.Frame(self,padding=20)
        content_frame.grid(column=0, row=0, sticky=(N, W, E, S))
        content_frame.grid_rowconfigure(0,weight=1)
        content_frame.grid_columnconfigure(0,weight=1)

        title_label = ttk.Label(content_frame,text="Add Person Info")
        title_label.grid(row=0,column=0)

        ####### Mandatory Info
        mandatory_frame = ttk.Labelframe(content_frame,text="Mandatory Info")
        mandatory_frame.grid(column=0,row=1,sticky=(N,W,E,S))

        #ask for given name
        given_name_label = ttk.Label(mandatory_frame,text="Given Name: ")
        given_name_label.grid(row=1,column=0)
        self._given_name = tk.StringVar()
        self.given_name_entry = ttk.Entry(mandatory_frame,textvariable=self._given_name)
        self.given_name_entry.grid(row=1,column=1)

        #for highlighting missing fields and turning off red when user is inputting
        self.given_name_entry.bind("<FocusOut>",lambda event: style_guide.highlight_mandatory_entry_on_focus_out(self.given_name_entry))
        self.given_name_entry.bind("<FocusIn>",lambda event: style_guide.highlight_mandatory_entry_on_focus_in(self.given_name_entry))

        #ask for last name
        last_name_label = ttk.Label(mandatory_frame,text="Last Name: ")
        last_name_label.grid(row=2,column=0)
        self._last_name = tk.StringVar()
        self.last_name_entry = ttk.Entry(mandatory_frame,textvariable=self._last_name)
        self.last_name_entry.grid(row=2,column=1)

        #for highlighting missing fields and turning off red when user is inputting
        self.last_name_entry.bind("<FocusOut>",lambda event: style_guide.highlight_mandatory_entry_on_focus_out(self.last_name_entry))
        self.last_name_entry.bind("<FocusIn>",lambda event: style_guide.highlight_mandatory_entry_on_focus_in(self.last_name_entry ))

        #########Optional
        #optional arguments: ORCID and affiliations
        optional_frame = ttk.Labelframe(content_frame,text="Optional")
        optional_frame.grid(column=0,row=2,sticky=(N,W,E,S))
        
        affiliation_label = ttk.Label(optional_frame,text="Affiliation(s): ")
        affiliation_label.grid(row=3,column=0)
        self._affiliation = tk.StringVar()
        affiliation_entry = ttk.Entry(optional_frame,textvariable=self._affiliation)
        affiliation_entry.grid(row=3,column=1)

        self._id_type = tk.StringVar()
        id_type_combobox = ttk.Combobox(optional_frame,
                                        textvariable=self._id_type,
                                        values=('ORCID',),
                                        state="readonly")
        id_type_combobox.grid(row=4,column=0)
        self.after(1, lambda: id_type_combobox.current(0))
        self._id = tk.StringVar()
        id_entry = ttk.Entry(optional_frame,textvariable=self._id)
        id_entry.grid(row=4,column=1)

        submit_button = ttk.Button(content_frame,text="Submit",command=lambda: self._on_submit(submit_function))
        submit_button.grid(row=10,column=0,pady=10)

    def _on_submit(self,submit_function: Callable):
        """Wrapper enforcing validation before submit callable is executed

        Args:
            submit_function (Callable): Callable to be executed upon hitting submit
        """
        #enforce mandatory fields
        #don't return immediately so all incomplete fields highlighted together
        is_first_name_bad = False
        is_last_name_bad = False
        if not self._given_name.get():
            style_guide.show_error_in_entry(self.given_name_entry)
            is_first_name_bad = True
        
        if not self._last_name.get():
            style_guide.show_error_in_entry(self.last_name_entry)
            is_last_name_bad = True
        
        is_exit = is_first_name_bad or is_last_name_bad
        #exit now
        if is_exit:
            return
        
        submit_function()

    @property
    def given_name(self):
        return self._given_name.get()

    @property
    def last_name(self):
        return self._last_name.get()

    @property
    def affiliation(self):
        return self._affiliation.get()

    @property
    def id_type(self):
        return self._id_type.get()
    
    @property
    def id(self):
        return self._id.get()

class CreateAuthorOrgWindow(Toplevel):
    """Window containing form to create author (organisation)

    Args:
        Toplevel (_type_): tkinter Window object
    """
    def __init__(self,root, submit_function: Callable):
        """
        Initializes the CreateAuthorWindow.
        Args:
            root (Tk): The root window of the Tkinter application.
            submit_function (Callable): The function to be called when the submit button is pressed.
        """
        super().__init__(root)

        self.title("Specify Author Information") 
        self.grid_rowconfigure(0,weight=1)
        self.grid_columnconfigure(0,weight=1)

        content_frame = ttk.Frame(self,padding=20)
        content_frame.grid(column=0, row=0, sticky=(N, W, E, S))
        content_frame.grid_rowconfigure(0,weight=1)
        content_frame.grid_columnconfigure(0,weight=1)

        title_label = ttk.Label(content_frame,text="Add Organisation Info")
        title_label.grid(row=0,column=0)

        ####### Mandatory Info
        mandatory_frame = ttk.Labelframe(content_frame,text="Mandatory Info")
        mandatory_frame.grid(column=0,row=1,sticky=(N,W,E,S))

        #ask for org name
        name_label = ttk.Label(mandatory_frame,text="Name: ")
        name_label.grid(row=1,column=0)
        self._org_name = tk.StringVar()
        self.name_entry = ttk.Entry(mandatory_frame,textvariable=self._org_name)
        self.name_entry.grid(row=1,column=1)

        #for highlighting missing fields and turning off red when user is inputting
        self.name_entry.bind("<FocusOut>",lambda event: style_guide.highlight_mandatory_entry_on_focus_out(self.name_entry))
        self.name_entry.bind("<FocusIn>",lambda event: style_guide.highlight_mandatory_entry_on_focus_in(self.name_entry))

        #########Optional
        #optional arguments: ORCID and affiliations
        optional_frame = ttk.Labelframe(content_frame,text="Optional")
        optional_frame.grid(column=0,row=2,sticky=(N,W,E,S))
        
        affiliation_label = ttk.Label(optional_frame,text="Affiliation(s): ")
        affiliation_label.grid(row=3,column=0)
        self._affiliation = tk.StringVar()
        affiliation_entry = ttk.Entry(optional_frame,textvariable=self._affiliation)
        affiliation_entry.grid(row=3,column=1)

        self._id_type = tk.StringVar()
        id_type_combobox = ttk.Combobox(optional_frame,
                                        textvariable=self._id_type,
                                        values=('ORCID',),
                                        state="readonly")
        id_type_combobox.grid(row=4,column=0)
        self.after(1, lambda: id_type_combobox.current(0))
        self._id = tk.StringVar()
        id_entry = ttk.Entry(optional_frame,textvariable=id)
        id_entry.grid(row=4,column=1)

        submit_button = ttk.Button(content_frame,text="Submit",command=lambda: self._on_submit(submit_function))
        submit_button.grid(row=10,column=0,pady=10)

    def _on_submit(self,submit_function: Callable):
        """Wrapper enforcing validation before submit callable is executed

        Args:
            submit_function (Callable): Callable to be executed upon hitting submit
        """
        #enforce mandatory fields, show missing field
        if not self._org_name.get():
            style_guide.show_error_in_entry(self.name_entry)
            return
        
        submit_function()

    @property
    def name(self):
        return self._org_name.get()

    @property
    def affiliation(self):
        return self._affiliation.get()

    @property
    def id_type(self):
        return self._id_type.get()
    
    @property
    def id(self):
        return self._id.get()


import ttkbootstrap as ttk
import tkinter as tk
from ttkbootstrap.constants import *
from ttkbootstrap.scrolled import ScrolledFrame
from typing import List,Dict,Tuple
from .upload_widgets.geometry import UploadGeometryFrame
from .upload_widgets.session import UploadSessionFrame
from .upload_widgets.output import UploadOutputFrame
from .upload_widgets.basic import UploadInfoFrame
from NekUpload.frontend.components.settings_manager import SettingsManager
from NekUpload.metadata.metadata import InvenioMetadata
from NekUpload.metadata.identifier import Identifier,IdentifierType
from NekUpload.metadata.user import InvenioOrgInfo,InvenioPersonInfo
import logging
from NekUpload.manager import NekManager,GeometryManager,SessionManager,OutputManager
from NekUpload.upload.invenio_db import InvenioRDM
from NekUpload.frontend import style_guide
from NekUpload.metadata.data_types import ResourceType

class UploadScene(ScrolledFrame):
    def __init__(self,root,parent,setting_manager: SettingsManager):
        super().__init__(parent,autohide=True)

        self.root = root
        self.setting_manager = setting_manager#contains settings data

        self.rowconfigure(0,weight=1)
        self.rowconfigure(1,weight=1)
        self.rowconfigure(2,weight=1)
        self.columnconfigure(0,weight=1)

        about_section: ttk.Frame = self._add_upload_info_section(self)
        about_section.grid(row=0,column=0,sticky=(NSEW))

        self.basic_info_section= UploadInfoFrame(self,self,self.setting_manager)
        self.basic_info_section.grid(row=1,column=0,sticky=NSEW,padx=10,pady=5)
        self.geometry_section = UploadGeometryFrame(self)
        self.geometry_section.grid(row=2,column=0,sticky=(NSEW),padx=10,pady=5)
        self.input_section = UploadSessionFrame(self)
        self.input_section.grid(row=3,column=0,sticky=(NSEW),padx=10,pady=5)
        self.output_section = UploadOutputFrame(self)
        self.output_section.grid(row=4,column=0,sticky=(NSEW),padx=10,pady=5)

        self.bind("<Configure>", self.update_wraplength)

        #upload button
        submit_button = ttk.Button(
            master=self,
            bootstyle=PRIMARY,
            text="Upload Datasets",
            command=self._upload_datasets
        )
        submit_button.grid(row=10,column=0,columnspan=10,sticky=NSEW,padx=10,pady=5)

    def update_wraplength(self, event):
        # Dynamically set the wraplength based on the width of the parent frame
        # Subtract a little for padding and margin
        self.upload_description.config(wraplength=event.width - 20)
        pass

    def _add_upload_info_section(self,parent) -> ttk.Frame:

        frame = ttk.Frame(master=parent)

        # Create the label for the title
        self.upload_info_label = ttk.Label(
            master=frame,
            text="Uploading Nektar++ Datasets",
            font=("TkDefaultFont", 20, "bold", "underline"),
            anchor="w",
            bootstyle=PRIMARY
        )
        self.upload_info_label.grid(row=0, column=0, pady=5, sticky=W)

        # Create the description label
        self.upload_description = ttk.Label(
            master=frame,
            text=("A Nektar++ dataset consists of: \n\n"
                " - Geometry Files\n"
                " - Input Files\n"
                " - Output Files\n"
                "\n"
                "There are currently two ways of uploading. The traditional way is that you have all the geometry files, "
                "input files and output files to be uploaded. Another way is that the geometry file already exists in the "
                "database, and you wish to link your input and ouptut files against it. This prevents repeated instances of "
                "same geometry file."),
            font=("TKDefaultFont", 12),
            anchor="w",
            justify="left",
        )
        self.upload_description.grid(row=1, column=0, pady=10, sticky="nsew")

        return frame

    def _check_dataset_inputs(self) -> bool:
        """Check all entry data is present, if not change style

        Returns:
            bool: _description_
        """

        is_error = True

        #check geometry widgets
        if not self.geometry_section.geometry_dataset_title:
            is_error = False

        if not self.geometry_section.geometry_file_name:
            is_error = False

        self.geometry_section.add_error_style_to_mandatory_entries()

        #check input widgets
        if not self.input_section.session_dataset_title:
            is_error = False

        if not self.input_section.session_file_name:
            is_error = False
        
        self.input_section.add_error_style_to_mandatory_entries()
        
        #check output widgets 
        if not self.output_section.output_dataset_title:
            is_error = False

        if not self.output_section.output_file_name:
            is_error = False

        self.output_section.add_error_style_to_mandatory_entries()

        return is_error
    
    def _upload_datasets(self):
        try:
            is_all_info_entered = True

            if not self.basic_info_section.author_list:
                logging.error("No authors entered. Please add authors.")
                is_all_info_entered = False

            if not self._check_dataset_inputs():
                logging.error("Missing mandatory dataset inputs. Please see red highlighted entries.")
                is_all_info_entered = False

            #exit if mandatory inputs not filled
            if not is_all_info_entered:
                return

            logging.info("UPLOADING...")

            #get general info
            publication_date: str = self.basic_info_section.publication_date_iso
            COMMUNITY_SLUG: str = self.basic_info_section.community_slug
            URL: str = self.setting_manager.database_url

            #ASSIGN USERS
            author_list: List[InvenioOrgInfo | InvenioPersonInfo] = []
            for author in self.basic_info_section.author_list:
                author_info: InvenioOrgInfo | InvenioPersonInfo = None
                
                if author["type"] == "personal":
                    given_name = author["given_name"]
                    last_name = author["last_name"]
                    author_info = InvenioPersonInfo(given_name,last_name)
                elif author["type"] == "organizational":
                    name = author["name"]
                    author_info = InvenioOrgInfo(name)

                if author["id"] != "":
                    mapping = {"ORCID" : IdentifierType.ORCID}
                    id = Identifier(author["id"],mapping[author["id_type"]])
                    author_info.add_identifier(id)

                #also affiliations, but not supported in backend for now
                author_list.append(author_info)

            #titles for each dataset
            geometry_title = self.geometry_section.geometry_dataset_title
            input_title = self.input_section.session_dataset_title
            output_title = self.output_section.output_dataset_title

            #create metadata for each one
            metadata_geometry = InvenioMetadata(geometry_title,publication_date,author_list,ResourceType.PHYSICAL_OBJECT)
            metadata_input = InvenioMetadata(input_title,publication_date,author_list,ResourceType.DATASET)
            metadata_output = InvenioMetadata(output_title,publication_date,author_list,ResourceType.DATASET)

            metadata_geometry.add_publisher("NekRDM")
            metadata_input.add_publisher("NekRDM")
            metadata_output.add_publisher("NekRDM")

            if geometry_description := self.geometry_section.geometry_description:
                metadata_geometry.add_description(geometry_description)

            #get files
            geometry_file: str = self.geometry_section.geometry_file_name
            session_file: str = self.input_section.session_file_name
            output_file: str = self.output_section.output_file_name

            geometry_optional_files: List[str] = self.geometry_section.geometry_optional_files

            #use geometry title for now
            #TODO in future, look to splinter records based on geometry, input, output        
            geometry_uploader = GeometryManager(geometry_file,geometry_optional_files,metadata_geometry,InvenioRDM())
            input_uploader = SessionManager(session_file,[],metadata_input,InvenioRDM())
            output_uploader = OutputManager(output_file,metadata=metadata_output,uploader=InvenioRDM())

            manager = NekManager(geometry_uploader,input_uploader,output_uploader)
        
            manager.validate()
            manager.execute_upload(URL,
                            self.setting_manager.token,
                            COMMUNITY_SLUG)
        except Exception as e:
            logging.error(e)


class HDF5ReaderException(Exception):
    """_summary_

    Args:
        Exception (_type_): _description_
    """
    def __init__(self, message):
        super().__init__(message)

class XMLReaderException(Exception):
    """_summary_

    Args:
        Exception (_type_): _description_
    """
    def __init__(self, message):
        super().__init__(message)


from . import invenio_rdm_api as invenioAPI 
from typing import List, Dict, Any
import requests
import logging
import os

# Note that all private functions will return requests.Response for flexibility and testing purposes
# Each member function is responsible for handling that requests.Response and parsing relevant data, stored in class files

#TODO Need to refactor
# 1. Relying on side effects to instantiate class is annoying
# 2. Keep the upload method? and also have builder methods if users want to customise the upload process?
# 3. In which case, need way of returning necessary information?
class InvenioRDM:
    """Responsible for InvenioRDM database upload logic.
    """
    def __init__(self) -> None:
        """Class initialiser
        """
        #instantiated after draft record is created
        self.record_id: str = None

        #community uuid extracted from get_community
        self.community_uuid: str = None

        #acquired after submitting request to community
        self.request_id: str = None

        #acquired after reserving doi during draft creation 
        self.doi: str=None
        self.record_link: str=None
        self.self_link: str=None

    def upload_files(self,url: str, token: str, file_paths: List[str], metadata: Dict[str,Any],community_id: str) -> None:  
        """Upload files to an InvenioRDM repository and submit to community for review. Full workflow.

        Args:
            url (str): Base URL to the InvenioRDM repository.  For example: "https://my-invenio-rdm.example.com"            
            token (str): User access token
            file_paths (List[str]): List of paths of files to be uploaded
            metadata (Dict[str,Any]): Metadata to be uploaded
            community_id (str): Invenio community id for upload or slug
        """        
        #prevent mixup of files from previous uploads
        self._clear()

        self.create_record(url,token,file_paths,metadata)
        self.submit_to_community(url,token,community_id)

    def create_record(self,url: str, token: str, file_paths: List[str], metadata: Dict[str,Any]):
        """Create a draft record with the files

        Args:
            url (str): _description_
            token (str): _description_
            file_paths (List[str]): _description_
            metadata (Dict[str,Any]): _description_
        """
        #create the draft
        create_record_response = invenioAPI.create_draft_record(url,token,metadata)
        self._handle_create_draft_response(create_record_response)
        logging.info(f"Record draft {self.record_id} has been created.")

        #after creation of draft, if any api calls fail, clean up the draft to prevent clutter
        try: 
            #set up the file locations in the record
            file_name_list = [self._get_file_name(file) for file in file_paths]
            invenioAPI.prepare_file_upload(url,token,self.record_id,file_name_list) #does batch
            logging.info(f"Draft {self.record_id} now ready for file uploads.")

            doi_response = invenioAPI.reserve_doi_draft(url,token,self.record_id)
            self._handle_doi_response(doi_response)

            #upload each file and commit
            for file,filename in zip(file_paths,file_name_list):
                invenioAPI.upload_file(url,token,self.record_id,file)
                invenioAPI.commit_file_upload(url,token,self.record_id,filename)
                logging.info(f"File {file} has been uploaded and committed to record {self.record_id}")

        except Exception as e:
            invenioAPI.delete_draft(url,token,self.record_id)
            logging.info(f"Record draft {self.record_id} has been deleted due to error: {e}")
            raise

    def submit_to_community(self,url:str,token:str,community_id:str):
        """Submit draft record for community review

        Args:
            url (str): _description_
            token (str): _description_
            community_id (str): _description_
        """
        #now try submitting to community, if fails here, ask user to manuallyu submit to community
        try:
            #get community uuid, then submit record to communtiy for review
            community_request = invenioAPI.get_community(url,token,community_id)
            self._handle_community_response(community_request)
            create_review_request_response = invenioAPI.submit_record_to_community(url,token,self.community_uuid,self.record_id)
            self._handle_create_review_request_response(create_review_request_response)

            payload = {
                "content": "This record was submitted via the Nektar++ validation and upload pipeline",
                "format": "html"
            } #this is a comment for the reviewer as to what this record is
            
            invenioAPI.submit_record_for_review(url,token,self.record_id,payload)
            logging.info(f"Record {self.record_id} has been submitted to community {self.community_uuid} for review")
        except Exception as e:
            logging.info(f"Failed to submit to community, please manually attempt on IvenioRDM, due to error: {e}")
            raise

    def _get_community_uuid(self) -> str:
        """Get community UUID

        Returns:
            str: Community UUID
        """
        return self.community_uuid

    def _get_file_name(self,file_path: str) -> str:
        """Given a file path, extract the file name

        Args:
            file_path (str): Path to file

        Returns:
            str: File name
        """
        file_name = os.path.basename(file_path)
        return file_name

    def _handle_create_draft_response(self,response: requests.Response) -> None:
        """Handles response from creation of draft. Reads useful information into class variables

        Args:
            response (requests.Response): Create draft response
        """
        data = response.json()
        self.record_id = data["id"]

    def _handle_community_response(self,response: requests.Response) -> None:
        """Handles get community request response

        Args:
            response (requests.Response): Get Community response
        """
        data = response.json()
        self.community_uuid = data["id"]

    def _handle_doi_response(self,response: requests.Response) -> None:
        """Handles the doi reservation request response

        Args:
            response (requests.Response): DOI reservation request response
        """
        data: Dict[str,Any] = response.json()
        
        self.doi = data["pids"]["doi"]["identifier"]
        self.record_link = data["links"]["record_html"]
        self.self_link = data["links"]["self_html"]

    def _handle_create_review_request_response(self,response: requests.Response) -> None:
        """Handles create review request response

        Args:
            response (requests.Response): Create review request response
        """
        data = response.json()
        self.request_id = data["id"]

    def _clear(self) -> None:
        """Reset internal state of the invenioRDM uploader
        """
        self.record_id = None
        self.community_uuid = None
        self.request_id = None
        self.doi = None
        self.record_link = None
        self.self_link = None


import ttkbootstrap as ttk
from ttkbootstrap.constants import *

class ExploreScene(ttk.Frame):
    def __init__(self,parent):
        super().__init__(parent)

        self.rowconfigure(0,weight=1)
        self.rowconfigure(1,weight=1)
        self.rowconfigure(2,weight=1)
        self.columnconfigure(0,weight=1)
        self.columnconfigure(1,weight=1)
        self.columnconfigure(2,weight=1)

        self.placeholder = ttk.Label(
            master=self,
            text="EXPLORE HERE",
            font=("TkDefaultFont", 20),
            anchor="center"
        )

        self.placeholder.grid(row=1,column=1,sticky=(NSEW))
        


from enum import Enum,auto

class BasisType(Enum):
    """Enum denoting what basis type is used in the construction of spectral elements in Nektar++.
    As defined in Nektar/library/LibUtilities/Foundations/BasisType.h
    """
    #Must be defined exactly as I/O files use integer values to reference these points
    NO_BASIS_TYPE = 1
    ORTHO_A = 2     # Principle Orthogonal Functions
                    # \f$\widetilde{\psi}^a_p(z_i)\f$
    ORTHO_B = 3     # Principle Orthogonal Functions
                    # \f$\widetilde{\psi}^b_{pq}(z_i)\f$
    ORTHO_C = 4     # Principle Orthogonal Functions
                    # \f$\widetilde{\psi}^c_{pqr}(z_i)\f$
    MODIFIED_A = 5  # Principle Modified Functions \f$ \phi^a_p(z_i) \f$
    MODIFIED_B = 6  # Principle Modified Functions \f$ \phi^b_{pq}(z_i) \f$
    MODIFIED_C = 7  # Principle Modified Functions \f$ \phi^c_{pqr}(z_i) \f$
    ORTHO_PYR_C = 8 # Principle Orthogonal Functions
                    # \f$\widetilde{\psi}^c_{pqr}(z_i) for Pyramids\f$
    MODIFIED_PYR_C = 9# Principle Modified Functions \f$ \phi^c_{pqr}(z_i) for
                    # Pyramids\f$
    FOURIER = 10    # Fourier Expansion \f$ \exp(i p\pi  z_i)\f$
    GLL_LAGRANGE = 11# Lagrange for SEM basis \f$ h_p(z_i) \f$
    GAUSS_LAGRANGE = 12# //!< Lagrange Polynomials using the Gauss points \f$
                    #!< h_p(z_i) \f$
    LEGENDRE = 13   # Legendre Polynomials \f$ L_p(z_i) = P^{0,0}_p(z_i)\f$. Same
                    # as Ortho_A
    CHEBYSHEV = 14  # Chebyshev Polynomials \f$ T_p(z_i) =
                    # P^{-1/2,-1/2}_p(z_i)\f$
    MONONMIAL = 15  # Monomial polynomials \f$ L_p(z_i) = z_i^{p}\f$
    FOURIER_SINGLE_MODE = 16    # Fourier ModifiedExpansion with just the first mode
                                #\f$ \exp(i \pi  z_i)\f$
    FOURIER_HALF_MODE_RE = 17  #//!< Fourier Modified expansions with just the real part
                                #//!< of the first mode  \f$ Re[\exp(i \pi  z_i)]\f$
    FOURIER_HALF_MODE_IM = 18  #//!< Fourier Modified expansions with just the imaginary
                                #//!< part of the first mode  \f$ Im[\exp(i \pi  z_i)]\f$
    SIZE_BASIS_TYPE = 19        #//!< Length of enum list    

class Elements(Enum):
    """Enum denoting element shape and type.
    """
    VERT = 1
    CURVE_NODE = 2
    SEG = 3
    CURVE_EDGE = 4
    TRI = 5
    QUAD = 6
    CURVE_FACE = 7
    HEX = 8
    PRISM = 9
    PYR = 10
    TET = 11
    COMPOSITE = 12
    DOMAIN = 13
    EDGE = 14
    FACE = 15

class ExpansionType(Enum):
    """Enum used for defining the expansion type of a spectral element in Nektar++.
    Same enums as defined in nektar/library/LibUtilities/Foundations/MeshGraph.h
    """
    #Must be defined exactly as I/O files use integer values to reference these points
    NO_EXPANSION_TYPE = 1           # No expansion type
    MODIFIED = 2                    # Modified expansion
    MODIFIED_QUAD_PLUS_1 = 3        # Modified quad plus 1
    MODIFIED_QUAD_PLUS_2 = 4        # Modified quad plus 2
    MODIFIED_GLL_RADAU_10 = 5       # Modified GLL Radau 10
    ORTHOGONAL = 6                  # Orthogonal expansion
    GLL_LAGRANGE = 7                # GLL Lagrange expansion
    GLL_LAGRANGE_SEM = 8            # GLL Lagrange SEM expansion
    GAUSS_LAGRANGE = 9              # Gauss Lagrange expansion
    GAUSS_LAGRANGE_SEM = 10         # Gauss Lagrange SEM expansion
    FOURIER = 11                    # Fourier expansion
    FOURIER_SINGLE_MODE = 12        # Fourier single mode expansion
    FOURIER_HALF_MODE_RE = 13       # Fourier half mode (real) expansion
    FOURIER_HALF_MODE_IM = 14       # Fourier half mode (imaginary) expansion
    CHEBYSHEV = 15                  # Chebyshev expansion
    FOURIER_CHEBYSHEV = 16          # Fourier-Chebyshev expansion
    CHEBYSHEV_FOURIER = 17          # Chebyshev-Fourier expansion
    FOURIER_MODIFIED = 18           # Fourier modified expansion
    EXPANSION_TYPE_SIZE = 19        # Size of the expansion type list

class IntegrationPoint(Enum):
    """Enum describing what integration/quadrature points are used for evaluation in spectral elements in Nektar++.
    Same enums as defined in nektar/library/LibUtilities/Foundations/PointsType.h
    """
    #Must be defined exactly as I/O files use integer values to reference these points
    GAUSS_GAUSS_LEGENDRE = 1                # 1D Gauss-Gauss-Legendre quadrature points
    GAUSS_RADAU_M_LEGENDRE = 2             # 1D Gauss-Radau-Legendre quadrature points, pinned at x=-1
    GAUSS_RADAU_P_LEGENDRE = 3             # 1D Gauss-Radau-Legendre quadrature points, pinned at x=1
    GAUSS_LOBATTO_LEGENDRE = 4             # 1D Gauss-Lobatto-Legendre quadrature points
    GAUSS_GAUSS_CHEBYSHEV = 5              # 1D Gauss-Gauss-Chebyshev quadrature points
    GAUSS_RADAU_M_CHEBYSHEV = 6            # 1D Gauss-Radau-Chebyshev quadrature points, pinned at x=-1
    GAUSS_RADAU_P_CHEBYSHEV = 7            # 1D Gauss-Radau-Chebyshev quadrature points, pinned at x=1
    GAUSS_LOBATTO_CHEBYSHEV = 8            # 1D Gauss-Lobatto-Chebyshev quadrature points
    GAUSS_RADAU_M_ALPHA0_BETA1 = 9         # Gauss Radau pinned at x=-1, alpha=0, beta=1
    GAUSS_RADAU_M_ALPHA0_BETA2 = 10        # Gauss Radau pinned at x=-1, alpha=0, beta=2
    GAUSS_RADAU_M_ALPHA1_BETA0 = 11        # Gauss Radau pinned at x=-1, alpha=1, beta=0
    GAUSS_RADAU_M_ALPHA2_BETA0 = 12        # Gauss Radau pinned at x=-1, alpha=2, beta=0
    GAUSS_KRONROD_LEGENDRE = 13            # 1D Gauss-Kronrod-Legendre quadrature points
    GAUSS_RADAU_KRONROD_M_LEGENDRE = 14    # 1D Gauss-Radau-Kronrod-Legendre quadrature points, pinned at x=-1
    GAUSS_RADAU_KRONROD_M_ALPHA1_BETA0 = 15 # 1D Gauss-Radau-Kronrod-Legendre pinned at x=-1, alpha=1, beta=0
    GAUSS_LOBATTO_KRONROD_LEGENDRE = 16    # 1D Lobatto Kronrod quadrature points
    POLY_EVENLY_SPACED = 17                # 1D Evenly-spaced points using Lagrange polynomial
    FOURIER_EVENLY_SPACED = 18             # 1D Evenly-spaced points using Fourier Fit
    FOURIER_SINGLE_MODE_SPACED = 19        # 1D Non Evenly-spaced points for Single Mode analysis
    BOUNDARY_LAYER_POINTS = 20             # 1D power law distribution for boundary layer points
    BOUNDARY_LAYER_POINTS_REV = 21         # 1D power law distribution for boundary layer points (reversed)
    NODAL_TRI_ELEC = 22                    # 2D Nodal Electrostatic Points on a Triangle
    NODAL_TRI_FEKETE = 23                  # 2D Nodal Fekete Points on a Triangle
    NODAL_TRI_EVENLY_SPACED = 24           # 2D Evenly-spaced points on a Triangle
    NODAL_TET_EVENLY_SPACED = 25           # 3D Evenly-spaced points on a Tetrahedron
    NODAL_TET_ELEC = 26                    # 3D Nodal Electrostatic Points on a Tetrahedron
    NODAL_PRISM_EVENLY_SPACED = 27         # 3D Evenly-spaced points on a Prism
    NODAL_PRISM_ELEC = 28                  # 3D electrostatically spaced points on a Prism
    NODAL_TRI_SPI = 29                     # 2D Nodal Symmetric Positive Internal triangle (Whitherden, Vincent)
    NODAL_TET_SPI = 30                     # 3D Nodal Symmetric Positive Internal tetrahedron (Whitherden, Vincent)
    NODAL_PRISM_SPI = 31                   # 3D Prism SPI
    NODAL_QUAD_ELEC = 32                   # 2D GLL for quad
    NODAL_HEX_ELEC = 33                    # 3D GLL for hex
    SIZE_POINTS_TYPE = 34                  # Length of enum list    

class SolverType(Enum):
    """Enum describing the available solvers in Nektar++
    """
    ADR_SOLVER = auto()
    ACOUSTIC_SOLVER = auto()
    INCOMPRESSIBLE_NAVIER_STOKES_SOLVER = auto()
    CARDIAC_EPS_SOLVER = auto()
    COMPRESSIBLE_FLOW_SOLVER = auto()
    DIFFUSION_SOLVER = auto()
    IMAGE_WARPING_SOLVER = auto()
    LINEAR_ELASTIC_SOLVER = auto()
    MMF_SOLVER = auto()
    PULSE_WAVE_SOLVER = auto()
    SHALLOW_WATER_SOLVER = auto()
    VORTEX_WAVE_INTERACTION_SOLVER = auto()